#!/usr/bin/env python
#----------------------------------------------------------------------------
# Name:         OritatamiSim.py
# Purpose:      A basic oritatami simulator
#
# Author:       Matt Patitz
#
# Created:      
# RCS-ID:       
# Copyright:    
# Licence:      
#----------------------------------------------------------------------------

"""
This program can load the definition of an oritatami system from a file and
simulate the growth of an assembly based on it.

Requirements:
    This application requires Python 2.x due to its dependency on wxPython

Dependencies:
    wxPython    (download installer for your platform here:  https://www.wxpython.org/download.php)
    pil         (via pillow, install using 'pip install pillow', pip can be found in your Python27\Scripts dir)
"""


"""
MJP TODO list:
    - fix parent-child links and bonds to either never or always cover beads
    - don't make path to system definition an absolute path
        - when creating a new system, make the assembly's path to the system definition relative?
    - CRASH: when bead cycle window is open during fast forward
    - when parameters DelayLength, Arity, LockWithoutBinding are set, should they be remembered when creating new?
        - test setting and saving
    - add saving of systems
        - implement flag to denote when a system has been edited to warn when (re)loading
        - when loading, don't allow whitespace in names
    - add functionality to window for typing and viewing primary structure
        - add help button
        - BUG: sometimes after reset (or reload?) the entire sequence becomes drawn in bold
        - disallow any whitespace entry
    - fix deletion of nodes when in drawing mode:
        - if a bead in the middle of a string is deleted, going forward is messed up
        - redrawing is very inefficient
        - deleting from the middle will cause additional problems when trying to insert back in...
    - terminate thread on exit if it's running!
        - doesn't get killed on breakpoint or when can't find free loc
    - on mouse over:  display number of addition (requires cache)
    - make help window a single instance which you can turn on and off
    - add option for toggling locking by only secondary structure of floppy segment
    - u: undo, r: reverse (will need to cache additions, allow seed change midway)
    - set icon in py2exe script so it will show up in task bar
"""

#  Potential variations of dynamics to consider:
#    - does last bead get fixed even without a connection?
#    - multiple bonds per bead
#    - full set of beads of delay length are transcribed and attach
#    - bonds are allowed to break for reconfiguration(?)

import wx
import wx.richtext as rt
import os
import random
import math
import re
import sys
import thread
from collections import deque
import itertools
import wx.aui
import time
import threading
import wx.lib.scrolledpanel as scrolled
from wx.lib.agw.shapedbutton import SButton, SBitmapButton
import wx.lib.agw.cubecolourdialog as CCD
import wx.html

RAND_SEED = -1

DocumentationFileList = [ "Overview", "System Parameters", "File Formats", "Tips and Tricks"]


GRIDSTARTSIZE = 30
GRIDGROWTHINC = 30
TRISIZE = 60

# IDs
ORIT_ID_FR             = 100
ORIT_ID_REV            = 101
ORIT_ID_STOP           = 102
ORIT_ID_FOR            = 103
ORIT_ID_FF             = 104
ORIT_ID_NEW            = 105
ORIT_ID_LOAD           = 106
ORIT_ID_SAVE           = 107
ORIT_ID_ZOOMIN         = 108
ORIT_ID_ZOOMOUT        = 109
ORIT_ID_MODE_PAN_ZOOM  = 110
ORIT_ID_MODE_DRAW_BEAD = 111
ORIT_ID_MODE_DRAW_LINK = 112
ORIT_ID_MODE_DRAW_BOND = 113
ORIT_ID_RESET          = 114
ORIT_ID_RESET_BEADSEQ  = 115
ORIT_ID_RELOAD_BEADSEQ = 116
ORIT_ID_CANCEL_BEADSEQ = 117
ORIT_ID_OK_NOCHECK     = 118
ORIT_ID_OK_CHECK       = 119
ORIT_ID_CANCEL         = 120

ORIT_MODE_PAN_ZOOM     = 0
ORIT_MODE_DRAW_BEAD    = 1
ORIT_MODE_DRAW_LINK    = 2
ORIT_MODE_DRAW_BOND    = 3

DEBUG_MSG = 0
MSG_MSG   = 1
IMP_MSG   = 2
WARN_MSG  = 3
ERROR_MSG = 4
DEBUG_MSG_LEVEL = WARN_MSG

# Result statuses for stepping forward in simulation
ORI_STAT_OK         = 0
ORI_STAT_NO_SPACE   = 1
ORI_STAT_NO_BINDING = 2
ORI_STAT_NONDET     = 3
ORI_STAT_MAX_FLOPPY = 4
ORI_STAT_ERROR      = 5


def PrintDebug(msg, level):
    if DEBUG_MSG_LEVEL <= level:
        print msg

colours = [
    "BLACK",
    "BLUE",
    "BLUE VIOLET",
    "BROWN",
    "CYAN",
    "DARK GREY",
    "DARK GREEN",
    "GOLD",
    "GREY",
    "GREEN",
    "MAGENTA",
    "NAVY",
    "PINK",
    "RED",
    "SKY BLUE",
    "VIOLET",
    "YELLOW",
    ]

N = 0
E = 1
S = 2
W = 3
ALLDIRS = 4
DIRSTRS = ("N", "E", "S", "W")

SHOW_GRID_AND_COORDS = 0
SHOW_GRID_ONLY       = 1
SHOW_NO_GRID         = 2

assertMode = wx.PYAPP_ASSERT_DIALOG
##assertMode = wx.PYAPP_ASSERT_EXCEPTION

ONBRS = [(0,1),(1,0),(1,-1),(0,-1),(-1,0),(-1,1)]

globalApp = None

class ONodeType:
    def __init__(self, name, color, bondToTypes):
        self.name = name
        self.color = color
        self.bondToTypes = bondToTypes
        
    def __repr__(self):
        return self.name
    
    def __str__(self):
        return self.name + " (Binds to: " + str(self.bondToTypes) + ")"

    def Output(self):
        outStr = "['" + self.name + "', '" + str(self.color) + "', " + str(self.bondToTypes) + "]"
        return outStr

    def BondsToNodeType(self, name):
        if name is None:
            return False
        if not self.bondToTypes is None:
            for b in self.bondToTypes:
                if b == name:
                    return True
        return False

class ONode:

    def __init__(self, otype, coords=None, parent=None):
        self.type = otype
        self.coords = coords
        self._bonds = []
        self.parent = parent
        self.child = None

    def __str__(self):
        outStr = "ONode[coords " + str(self.coords) + ", type " + str(self.type) + "]"
        return outStr

    def __repr__(self):
        outStr = "ONode[" + str(self.coords) + "]"
        return outStr
    
    def NumBonds(self):
        return len(self._bonds)
    
    def GetNthBond(self, n):
        return self._bonds[n]

    def AddBond(self, coords):
        if coords in self._bonds:
            return -1
        self._bonds.append(coords)
        return len(self._bonds)
    
    def RemoveBond(self, coords):
        if coords not in self._bonds:
            return -1
        self._bonds.remove(coords)
        return len(self._bonds)

class ONodeGrid:
    def __init__(self, width, height):
        self.MakeGrid(width, height)
        self.delayLength = 2    # Default delay of 2
        self.arity = 1          # Maximum number of bonds for a single bead
        self.lockWithoutBinding = False
        self.bindFullFloppySegment = False
        self.maxFloppyLength = 9    # Maximum length of a floppy segment to test during simulation before terminating
        self.seed = []
        self.beadCycle = []
        self.origBeadCycle = []
        self.currCyclePos = 0
        self.lastBeadLoc = None
        self.lastSeedLoc = None
        self.floppySegment = deque([])
        self.ONodeTypes = dict()
        self._numBeads = 0

    def GetNumBeads(self):
        return self._numBeads

    def GetWidth(self):
        return self._width
    
    def GetHeight(self):
        return self._height

    def MakeGrid(self, width, height, bResetBeadNum=False):
        self._width = width
        self._height = height
        self._grid = [[None for x in range(width)] for y in range(height)]
        self.minX = 0 - int(math.floor(width/2))
        self.maxX = int(math.floor(width/2))
        self.minY = 0 - int(math.floor(height/2))
        self.maxY = int(math.floor(height/2))
        if (width % 2) == 0:
            self.minX = self.minX + 1
        if (height % 2) == 0:
            self.minY = self.minY + 1
        if bResetBeadNum == True:
            self._numBeads = 0

    def GetLoc(self, coords):
        if coords[0] < self.minX or coords[0] > self.maxX or coords[1] < self.minY or coords[1] > self.maxY:
            print("Growing grid because of access to " + str(coords))
            self.GrowGrid()
        return self._grid[coords[0]-self.minX][coords[1]-self.minY]
    
    def SetLoc(self, coords, onode):
        if coords[0] < self.minX or coords[0] > self.maxX or coords[1] < self.minY or coords[1] > self.maxY:
            self.GrowGrid()
        self._grid[coords[0]-self.minX][coords[1]-self.minY] = onode
        return True

    def AddBond(self, coords1, coords2):
        loc1 = self.GetLoc(coords1)
        loc2 = self.GetLoc(coords2)
        if loc1 is None or loc2 is None:
            if loc1 is None:
                print("ERROR: trying to create a bond between " + str(coords1) + " and " + str(coords2) + " but " + str(coords1) + " is empty")
            else:
                print("ERROR: trying to create a bond between " + str(coords1) + " and " + str(coords2) + " but " + str(coords2) + " is empty")
            return False
        if loc1.AddBond(coords2) > -1:
            if loc2.AddBond(coords1) > -1:
                return True
        return False
    
    def RemoveBond(self, coords1, coords2):
        if self.GetLoc(coords1).RemoveBond(coords2) > -1:
            if self.GetLoc(coords2).RemoveBond(coords1) > -1:
                return True
        return False

    def GrowGrid(self):
        oldgrid = self._grid
        oldwidth = self._width
        oldheight = self._height
        oldminx = self.minX
        oldminy = self.minY
        self.MakeGrid(oldwidth + GRIDGROWTHINC, oldheight + GRIDGROWTHINC)
        for i in range(oldwidth):
            for j in range(oldheight):
                oldnode = oldgrid[i][j]
                self.SetLoc((i + oldminx, j + oldminy), oldnode)

    def AddNode(self, otype, coords, parentCoords, bWarnOnNoParent=True):
        childNode = ONode(otype, coords)
        if not self.GetLoc(coords) is None:
            print("ERROR: Trying to add another node at location " + str(coords))
            return
        self.SetLoc(coords, childNode)
        self.SetParent(coords, parentCoords, bWarnOnNoParent)
        self.lastBeadLoc = coords
        self._numBeads = self._numBeads + 1
        
    def SetParent(self, coords, parentCoords, bWarnOnNoParent=True):
        childNode = self.GetLoc(coords)
        if self.GetLoc(coords) is None:
            print("ERROR: Trying to set parent of non-existent node at location " + str(coords))
            return
        parentNode = None
        if (not parentCoords is None):
            parentNode = self.GetLoc(parentCoords)
        elif bWarnOnNoParent == True:
            print("Adding a node with no parent at location " + str(coords))
        if (not parentNode is None) and (not parentNode.child is None):
            print("Warning: Parent node already has a child, so not setting parent")
        else:
            childNode.parent = parentNode
            if not parentNode is None:
                parentNode.child = childNode
        
    def RemoveNode(self, coords):
        node = self.GetLoc(coords)
        if not node is None:
            if node.parent is not None:
                node.parent.child = None
                if self.lastBeadLoc == coords:
                    self.lastBeadLoc = node.parent.coords
            elif self.lastBeadLoc == coords:
                self.lastBeadLoc = None
            if node.child is not None:
                node.child.parent = None
            bonds = node._bonds #MJP: abusing private member here...
            for b in bonds:
                self.RemoveBond(coords, b)
            self.SetLoc(coords, None)
            self._numBeads = self._numBeads - 1
        else:
            print("ERROR: Trying to remove nonexistent node at location" + str(coords))


    def AreNbrLocs(self, coords1, coords2):
        if (coords1 is None) or (coords2 is None):
            return False
        for n in ONBRS:
            nbrCoords = (coords1[0]+n[0], coords1[1]+n[1])
            if nbrCoords == coords2:
                return True
        return False

    def GetNbrLocs(self, coords, bEmpty=False):
        retlist = []
        for n in ONBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            if bEmpty == False:
                retlist.append(nbrCoords)
            elif self.GetLoc(nbrCoords) is None:
                retlist.append(nbrCoords)
        return retlist

    def GetOccupiedNbrLocs(self, coords, bSkipParent=False, maxBonds=0):
        retlist = []
        for n in ONBRS:
            nbrCoords = (coords[0]+n[0], coords[1]+n[1])
            nbrNode = self.GetLoc(nbrCoords)
            if not nbrNode is None:
                if (bSkipParent == False) or (nbrNode.child is None) or (not nbrNode.child.coords == coords):
                    if nbrNode.NumBonds() <= maxBonds:
                        retlist.append(nbrNode)
        return retlist


    # Note: make sure beadsToAdd is reversed, i.e. the first bead to add is the last in the list!!!
    def GetPathsFreeToDistance(self, start, beadsToAdd):
        #print("GetPathsFreeToDistance called with start " + str(start) + ", beadsToAdd " + str(beadsToAdd))
        maxConns = 0
        maxConnsTo = []
        maxConnsFrom = []
        bestPaths = []

        nbrCoordSet = self.GetNbrLocs(start.coords, True)
        if len(nbrCoordSet) == 0:
            return (0, [[]], [[]], [[]])

        for nbrCoords in nbrCoordSet:
            
            poppedBead = beadsToAdd.popleft()
            newType = self.ONodeTypes[poppedBead]

            # We've found a free location to place the next bead
            nbrNode = ONode(newType, nbrCoords, start)
            
            # Temporarily add the bead at this location
            self.AddNode(poppedBead, nbrCoords, start.coords)

            # Going forward without binding it to anything
            if len(beadsToAdd) > 0:
                (conns, connsTo, connsFrom, paths) = self.GetPathsFreeToDistance(nbrNode, beadsToAdd)
                
                if conns >= maxConns:
                    for p in paths:
                        p.insert(0, nbrCoords)
                        #p.append(nbrCoords)

                if conns == maxConns:
                    maxConnsTo = maxConnsTo + connsTo
                    maxConnsFrom = maxConnsFrom + connsFrom
                    bestPaths = bestPaths + paths
                if conns > maxConns:
                    maxConns = conns
                    maxConnsTo = connsTo
                    maxConnsFrom = connsFrom
                    bestPaths = paths

            # For each bead that it can bind to, create that bond and go forward
            maxInnerConns=0
            maxInnerConnsTo=[]
            maxInnerConnsFrom=[]
            bestInnerPaths=[]

            nbrBound = False
            occupiedNbrs = self.GetOccupiedNbrLocs(nbrCoords, True, self.arity-1)
            #########################
            bondableNbrs = []
            for on in occupiedNbrs:
                if (newType.BondsToNodeType(on.type) == True):
                    bondableNbrs.append(on.coords)

            for i in range(1,self.arity+1):
                nbrSets = list(itertools.combinations(bondableNbrs, i))
                
                for nbrSet in nbrSets:
                    #print("for nbrCoords " + str(nbrCoords) + "nbrSet " + str(nbrSet))

                # See if nodes can bind and already existing node hadn't already bound to something (or is linked)
                #if (newType.BondsToNodeType(on.type) == True):
                    nbrBound = True
                    conns = 0
                    connsTo = [[]]
                    connsFrom = [[]]
                    paths = [[]]
                    
                    #self.AddBond(nbrCoords, on.coords)
                    for nbr in nbrSet:
                        self.AddBond(nbrCoords, nbr)

                    # Now go forward with the bond
                    if len(beadsToAdd) > 0:
                        (conns, connsTo, connsFrom, paths) = self.GetPathsFreeToDistance(nbrNode, beadsToAdd)
                    
                    #self.RemoveBond(nbrCoords, on.coords)
                    for nbr in nbrSet:
                        self.RemoveBond(nbrCoords, nbr)
                        for ct in connsTo:
                            ct.insert(0, nbrCoords)
                        for cf in connsFrom:
                            cf.insert(0, nbr)
                    conns = conns + len(nbrSet)
                    #===========================================================
                    # for ct in connsTo:
                    #     ct.insert(0, nbrCoords)
                    # for cf in connsFrom:
                    #     cf.insert(0, on.coords)
                    #===========================================================
                        
                    if conns >= maxInnerConns:
                        for p in paths:
                            p.insert(0, nbrCoords)
                            #p.append(nbrCoords)
                        
                    # Do a check for local max here
                    if conns == maxInnerConns:
                        maxInnerConnsTo = maxInnerConnsTo + connsTo
                        maxInnerConnsFrom = maxInnerConnsFrom + connsFrom
                        bestInnerPaths = bestInnerPaths + paths
                    if conns > maxInnerConns:
                        maxInnerConns = conns
                        maxInnerConnsTo = connsTo
                        maxInnerConnsFrom = connsFrom
                        bestInnerPaths = paths                    
            
            if nbrBound == False and len(beadsToAdd) == 0:  # Case where this is the last bead but couldn't bind to anything
                if maxConns == 0:
                    #bestPaths = list(bestPaths)
                    bestPaths.append([nbrCoords])
                    maxConnsTo.append([])
                    maxConnsFrom.append([])
            else:
                #print(" bestPaths after inner: " + str(bestPaths))
                if maxInnerConns == maxConns:
                    maxConnsTo = maxConnsTo + maxInnerConnsTo
                    maxConnsFrom = maxConnsFrom + maxInnerConnsFrom
                    bestPaths = bestPaths + bestInnerPaths
                if maxInnerConns > maxConns:
                    maxConns = maxInnerConns
                    maxConnsTo = maxInnerConnsTo
                    maxConnsFrom = maxInnerConnsFrom
                    bestPaths = bestInnerPaths
                #print(" bestPaths after combo: " + str(bestPaths))

            # Remove the bead that we temporarily added
            self.RemoveNode(nbrCoords)
            beadsToAdd.extendleft([poppedBead])
            
            #print(" bestPaths after loop iteration: " + str(bestPaths))

        #print("GetPathsFreeToDistance returning with bestPaths " + str(bestPaths))
        return (maxConns, maxConnsTo, maxConnsFrom, bestPaths)
        
    def InitializeFloppySegment(self):
        self.floppySegment = deque([])
        self.currentCyclePos = 0
        if len(self.beadCycle) == 0:
            globalApp.SetOutputMessage("WARNING: Bead cycle is empty!")
            return False
        for i in range(self.delayLength-1):
            self.floppySegment.append(self.beadCycle[i % len(self.beadCycle)])
            #self.currCyclePos = ((self.currCyclePos + 1) % len(self.beadCycle))
        print("In InitializeFloppySegment, floppy segment = " + str(self.floppySegment))

    def StepForward(self):
        retStat = ORI_STAT_OK
        retList = []

        if len(self.beadCycle) == 0:
            globalApp.ShowWarningDialog("WARNING: Bead cycle is empty!  Cannot simulate an assembly step.")
            globalApp.SetOutputMessage("WARNING: Bead cycle is empty!  Cannot simulate an assembly step.")
            return (ORI_STAT_NO_SPACE, [])
        
        self.floppySegment.append(self.beadCycle[(self.currCyclePos+len(self.floppySegment)) % len(self.beadCycle)])
        if self.bindFullFloppySegment == True:
            while len(self.floppySegment) < self.delayLength:
                self.floppySegment.append(self.beadCycle[(self.currCyclePos+len(self.floppySegment)) % len(self.beadCycle)])

        if self.maxFloppyLength < len(self.floppySegment):
            globalApp.ShowWarningDialog("The floppy segment has exceeded the maximum\nfloppy segment length setting (currently set to " + str(self.maxFloppyLength) + ").\nStopping simulation.")
            globalApp.SetOutputMessage("The floppy segment has exceeded the maximum floppy segment length setting (currently set to " + str(self.maxFloppyLength) + ").  Stopping simulation.")
            return (ORI_STAT_MAX_FLOPPY, [])

        PrintDebug("Testing addition of beads: " + str(self.floppySegment) + " starting from location " + str(self.lastBeadLoc), DEBUG_MSG)

        if self.lastBeadLoc is None:
            globalApp.ShowWarningDialog("There is no starting bead location.  Stopping simulation.")
            globalApp.SetOutputMessage("There is no starting bead location.  Stopping simulation.")
            return (ORI_STAT_NO_SPACE, [])
        startNode = self.GetLoc(self.lastBeadLoc)
        nextBeads = deque(self.floppySegment)
        #nextBeads.reverse()
        PrintDebug("About to call GetPathsFreeToDistance with startNode " + str(startNode) + " and nextBeads " + str(nextBeads), DEBUG_MSG)
        globalApp.SetOutputMessage("About to check possible configurations for: " + str(nextBeads))
        (maxNumBonds, maxConnsTo, maxConnsFrom, bestPaths) = self.GetPathsFreeToDistance(startNode, nextBeads)
        if (len(bestPaths) == 0) or (len(bestPaths[0]) == 0):
            globalApp.ShowWarningDialog("There are no free locations left to place beads.  Stopping simulation.")
            globalApp.SetOutputMessage("There are no free locations left to place beads.  Stopping simulation.")
            return (ORI_STAT_NO_SPACE, [])
        #cProfile.runctx('self.GetPathsFreeToDistance(startNode, 10, [])', None, locals())
        PrintDebug("Printing results from GetPathsFreeToDistance:", DEBUG_MSG)
        PrintDebug("maxConnsTo: " + str(maxConnsTo) + ", maxConnsFrom: " + str(maxConnsFrom) + ", bestPaths: " + str(bestPaths), DEBUG_MSG)
        PrintDebug("maxNumBonds: " + str(maxNumBonds), DEBUG_MSG)
        #for i in range(len(bestPaths)):
        #    print("best path: " + str(bestPaths[i]) + ", connsTo: " + str(maxConnsTo[i]) + ", connsFrom: " + str(maxConnsFrom[i]))

        if self.lockWithoutBinding == False and maxNumBonds == 0:
            PrintDebug("NO BINDING!!!!", DEBUG_MSG)
            globalApp.SetOutputMessage("No configurations with bonds found")
            return (ORI_STAT_NO_BINDING, [])
        else:
            pathsToAdd = []
            duplicateIndices = []
            count = 0
            for bp in bestPaths:
                path = []
                if self.bindFullFloppySegment == True:
                    numToAdd = len(self.floppySegment)
                    if len(bp) < numToAdd:
                        numToAdd = len(bp)
                    for i in range(numToAdd):
                        path.append(bp[i])
                else:
                    numToAdd = (len(self.floppySegment) - self.delayLength) + 1
                    if len(bp) < numToAdd:
                        numToAdd = len(bp)
                    for i in range(numToAdd):
                        path.append(bp[i])
                if not path in pathsToAdd:
                    pathsToAdd.append(path)
                else:
                    duplicateIndices.append(count)
                count = count + 1
            #print("Duplicate indices: " + str(duplicateIndices))
                
            loc = 0
            if len(pathsToAdd) > 1:
                globalApp.SetOutputMessage("Nondeterminism:  there are " + str(len(pathsToAdd)) + " paths with equal bonds: " )
                outStr = ""
                for outBP in pathsToAdd:
                    outStr = outStr + str(outBP)
                globalApp.SetOutputMessage(outStr)
                retStat = ORI_STAT_NONDET
                loc = random.randint(0, len(pathsToAdd)-1)
                globalApp.SetOutputMessage("Nondeterministically selected path " + str(loc) + ": " + str(pathsToAdd[loc]))
    
            parentLoc = self.lastBeadLoc
            pathToAdd = pathsToAdd[loc]
            
            # Check to see if this path is able to place all beads that it's supposed to
            bTooShort = False
            if self.bindFullFloppySegment == True:
                numToAdd = len(self.floppySegment)
                if len(pathToAdd) < numToAdd:
                    bTooShort = True
            else:
                numToAdd = (len(self.floppySegment) - self.delayLength) + 1
                if len(pathToAdd) < numToAdd:
                    bTooShort = True
            if bTooShort:
                globalApp.SetOutputMessage("There are not enough free locations left to place all beads.")
            
            for b in pathToAdd:
                newNodeType = self.floppySegment.popleft()
                retList.append(b)
                self.AddNode(newNodeType, b, parentLoc)
                self.currCyclePos = ((self.currCyclePos + 1) % len(self.beadCycle))
                parentLoc = b
                PrintDebug("Adding node of type " + newNodeType + " at " + str(b), DEBUG_MSG)
                PrintDebug("Floppy segment now " + str(self.floppySegment), DEBUG_MSG)
                
            adjLoc = loc
            for d in duplicateIndices:
                if d <= adjLoc:
                    adjLoc = adjLoc + 1
            for b in pathToAdd:
                # Check to see if the new node bonds to existing beads and add bond if so
                #MJP: currently going to lock in secondary structure
                for i in range(len(maxConnsTo[adjLoc])):
                    (x2,y2) = maxConnsTo[adjLoc][i]
                    (x3,y3) = maxConnsFrom[adjLoc][i]
                    #if ((x,y) == (x2,y2)) and ((x3,y3) not in bestPaths[adjLoc]):
                    if (b == (x2,y2)):
                        if self.AddBond(b, (x3,y3)) == False:
                            globalApp.ShowErrorDialog("ERROR: Failed to add bond between beads at " + str(b) + " and " + str((x3,y3)) +" \nStopping simulation.")
                            globalApp.SetOutputMessage("ERROR: Failed to add bond between beads at " + str(b) + " and " + str((x3,y3)) +" \nStopping simulation.")
                            return (ORI_STAT_ERROR, [])
        return (retStat, retList)

    def LoadOritatamiSystem(self, origFileName, origFileNameFullPath=None):
        fileName = origFileName
        if not origFileNameFullPath is None:
            fileName = origFileNameFullPath
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        try:
            os.stat(fileName)
            rFile = open(fileName, "r")
            print("Using oritatami system definition file '" + fileName + "'")
            lines = rFile.readlines()
            rFile.close()
    
            self.ONodeTypes = dict()
    
            for line in lines:
                if line.startswith("BeadTypes="):
                    try:
                        beadTypes = eval(line[len("BeadTypes="):])
                        for bead in beadTypes:
                            beadType = ONodeType(bead[0], bead[1], bead[2])
                            self.ONodeTypes[bead[0]] = beadType
                        self.FixBeadBondSymmetry(True)
                    except:
                        globalApp.ShowErrorDialog("Invalid bead type definitions.  See messages window.")
                        globalApp.SetOutputMessage("Invalid line '" + line + "'.  Bead types should be a comma separated list surrounded by '[' and ']'. Each bead type definition should similar to the following:  ['Name', 'Color', ['Name2', 'Name3']] where Name2 and Name3 are in a list of bead types that this type binds to.")
                        return False
                elif line.startswith("DelayLength="):
                    dlLine = line[len("DelayLength="):].strip()
                    self.delayLength = int(dlLine)
                    print("DelayLength: " + str(self.delayLength))
                    #if (self.delayLength > 2):
                        #MJP TODO: fix this issue!
                        #print("\n\nWARNING:  For delay longer than 2, newly extended beads could possibly bind to each other, but this is not yet implemented!!!\n")
                elif line.startswith("Arity="):
                    arityLine = line[len("Arity="):].strip()
                    self.arity = int(arityLine)
                    print("Arity: " + str(self.arity))
                elif line.startswith("LockWithoutBinding="):
                    lwbLine = line[len("LockWithoutBinding="):].strip()
                    self.lockWithoutBinding = False
                    if lwbLine == "True":
                        self.lockWithoutBinding = True
                    print("LockWithoutBinding: " + str(self.lockWithoutBinding))
                elif line.startswith("BindFullFloppySegment="):
                    bffsLine = line[len("BindFullFloppySegment="):].strip()
                    self.bindFullFloppySegment = False
                    if bffsLine == "True":
                        self.bindFullFloppySegment = True
                    print("BindFullFloppySegment:" + str(self.bindFullFloppySegment))
                elif line.startswith("BeadCycle="):
                    bcLine = line[len("BeadCycle="):].strip()
                    self.beadCycle = []
                    self.origBeadCycle = []
                    try:
                        self.beadCycle = eval(bcLine)
                        self.origBeadCycle = self.beadCycle
                    except:
                        globalApp.ShowErrorDialog("Invalid bead cycle definition.  See messages window.")
                        globalApp.SetOutputMessage("Invalid line '" + line + "'.  Bead cycle should be a comma separated list of bead type names surrounded by '[' and ']', similar to the following:  ['Name1', 'Name2', 'Name3'].")
                        return False
                    print("beadCycle: " + str(self.beadCycle))
    
            globalApp.menuSettings.Check(globalApp.itemSetLWB.GetId(), self.lockWithoutBinding)
            globalApp.menuSettings.Check(globalApp.itemSetBFFS.GetId(), self.bindFullFloppySegment)
            self.InitializeFloppySegment()
            globalApp.ResetBreakpoints()
            globalApp.systemDefinitionFile = origFileName
            globalApp.systemDefinitionFileFullPath = origFileNameFullPath
            globalApp.SetOutputMessage("Loaded bead type definitions from file '" + fileName + "'")
            print("Loaded bead set: " + str(self.ONodeTypes))
            return True;
        except:
            globalApp.ShowErrorDialog("\nERROR: Invalid oritatami system definition file: '" + fileName + "'\n")
            return False
        
    # Check for asymmetry in definitions of which bead types each can bind to
    def FixBeadBondSymmetry(self, bShowWarnings=True, setName=None, setList=None):
        retVal = False
        if not setName is None:
            btNames = self.ONodeTypes.keys()
            others = [item for item in btNames if item not in setList]
            for other in others:
                otherBT = self.ONodeTypes[other]
                if setName in otherBT.bondToTypes:
                    otherBT.bondToTypes.remove(setName)
                    retVal = True
        mismatches = []
        for btName in self.ONodeTypes:
            bt1 = self.ONodeTypes[btName]
            for bondable1 in bt1.bondToTypes:
                bt2 = self.ONodeTypes[bondable1]
                if btName not in bt2.bondToTypes:
                    mismatches.append((bt2.name, bt1.name))
        if len(mismatches) > 0:
            if bShowWarnings == True:
                globalApp.ShowWarningDialog("Invalid bead type definitions automatically repaired.  See messages window.")
            outStr = "The following pairs of bead types were only defined to bind in one of the bead type definitions.  They have been modified so both types include each other.\n"
            for mm in mismatches:
                outStr = outStr + mm[0] + " to " + mm[1] + "\n"
                self.ONodeTypes[mm[0]].bondToTypes.append(mm[1])
            if bShowWarnings == True:
                globalApp.SetOutputMessage(outStr)
            return True
        else:
            return retVal

    def LoadOritatamiAssembly(self, origFileName, bForceDefLoad=False):
        fileName = origFileName
        if len(fileName) < 1:
            return False
        elif os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        lastLoc = None
        try:
            os.stat(fileName)
            rFile = open(fileName, "r")
            print("Using oritatami assembly definition file '" + fileName + "'")
            lines = rFile.readlines()
            rFile.close()
            
            for line in lines:
                if line.startswith("OritatamiSystemDef="):
                    defFile = line[len("OritatamiSystemDef="):].strip()
                    if (bForceDefLoad == True) or (globalApp.systemDefinitionFile is None) or (not defFile == globalApp.systemDefinitionFile):
                        defFileFullPath = defFile
                        if os.path.isabs(defFileFullPath) == False:
                            dirPath = os.path.dirname(fileName)
                            defFileFullPath = os.path.join(dirPath, defFileFullPath)
                        if not self.LoadOritatamiSystem(defFile, defFileFullPath) == True:
                            return False
                elif line.startswith("Assembly="):
                    if self.LoadOritatamiAssemblyHelper(line) == False:
                        globalApp.SetOutputMessage("ERROR: Invalid oritatami assembly definition file: '" + fileName + "'")
                        return False
                elif line.startswith("LastBead="):
                    lastLocStr = line[len("LastBead="):].strip()
                    lastLoc = eval(lastLocStr)
        except:
            globalApp.ShowErrorDialog("\nERROR: Invalid oritatami assembly definition file: '" + fileName + "'")
            return False
        
        if not lastLoc is None:
            if self.GetLoc(lastLoc) is None:
                globalApp.SetOutputMessage("ERROR: Assembly definition has invalid last bead location " + str(lastLoc))
            else:
                self.lastSeedLoc = lastLoc
                self.lastBeadLoc = lastLoc
                
        globalApp.seedDefinitionFile = origFileName
        globalApp.SetOutputMessage("Loaded assembly definition from file '" + globalApp.seedDefinitionFile + "'")
        return True

    def LoadOritatamiAssemblyHelper(self, seedString):
        seedString = seedString[len("Assembly="):].strip()
        try:
            self.seed = []
            seedArr = eval(seedString)
            for b in seedArr:
                if b[0] in self.ONodeTypes:
                    self.seed.append(b)
                else:
                    globalApp.SetOutputMessage("ERROR: Trying to add a bead of undefined type '" + str(b[0]) + "' at " + str(b[1]))
        except:
            globalApp.ShowErrorDialog("\nERROR: Invalid oritatami assembly definition file.")
            return False
        return self.ResetToSeed()
        
    def ResetToSeed(self):
        print("Seed: " + str(self.seed))
        self.MakeGrid(self._width, self._height, True)
        savedLastSeedLoc = self.lastSeedLoc
        for b in self.seed:
            if b[0] in self.ONodeTypes:
                self.AddNode(b[0], b[1], None, False)
                self.lastSeedLoc = b[1]
            else:
                globalApp.SetOutputMessage("ERROR: Trying to add a bead of undefined type '" + str(b[0]) + "' at " + str(b[1]))
        if not savedLastSeedLoc is None:
            self.lastSeedLoc = savedLastSeedLoc
        # If a parent is not specified, the previous bead is used by default.  To specify no parent, use "None"
        parent = None
        for b in self.seed:
            if len(b) > 2:
                parent = b[2]
            self.SetParent(b[1], parent)
            parent = b[1]

        for b in self.seed:
            if len(b) > 3:
                blist = b[3]
                for bond in blist:
                    #MJP: check to see if these types should actually bond and warn if not!
                    t1 = self.GetLoc(b[1])
                    t2 = self.GetLoc(bond)
                    if (not t1 is None) and (not t2 is None):
                        if t1.type in self.ONodeTypes:
                            t1type = self.ONodeTypes[t1.type]
                            if t1type.BondsToNodeType(t2.type) == False:
                                globalApp.SetOutputMessage("WARNING: Creating a bond between non-binding bead types '" + t1.type + "' at " + str(b[1]) + " and '" + t2.type + "' at " + str(bond))
                        self.AddBond(b[1], bond)
        self.InitializeFloppySegment()
        self.lastBeadLoc = self.lastSeedLoc
        self.currCyclePos = 0
        globalApp.ResetBreakpoints()
        return True

#----------------------------------------------------------------------------

class MyCanvas(wx.ScrolledWindow):
    def __init__(self, parent, idIn, frame, size, app):
        wx.ScrolledWindow.__init__(self, parent, idIn, (0, 0), size=size, style=wx.SUNKEN_BORDER)

        self.frame = frame
        self.app = app
        self.x = self.y = 0
        self.scale = 1.0
        self.seed = {}
        self.origin = wx.Point()
        self.DrawQueue = []
        self.RemoveQueue = []
        self.objids = []
        self.dictIdToType = {}
        self.boundsdict = {}
        self.keepGoing = False
        self.showMouseOver = True
        self.breakpoints = []
        self.currentBreakpoint = -1
        self.breakOnNondeterminism = False
        self.confirmOverArity = True
        self.confirmInvalidBonds = True
        self._lock = threading.Lock()
        
        self.showGrid = SHOW_GRID_AND_COORDS
        
        self.mouseMode = ORIT_MODE_PAN_ZOOM
        self.drawingModeBeadType = None

        # create a PseudoDC to record our drawing
        self.pdc = wx.PseudoDC()
        self.pen_cache = {}
        self.brush_cache = {}
        self.gridPen = self.ColoredPen("BLACK", 1)
        self.axesPen = self.ColoredPen("RED", 1)
        self.bondPen = self.ColoredPen("RED", 2)
        self.linkPen = self.ColoredPen("BLACK", 4)
        self.grabbingCursorImg = wx.Image("images/grabbing.gif", wx.BITMAP_TYPE_GIF)

        self.SetBackgroundColour("WHITE")
        
        self.sideLen = TRISIZE
        self.triangleHeight = math.tan(math.radians(60)) * 0.5 * self.sideLen

        self.gridWidth = GRIDSTARTSIZE
        self.gridHeight = GRIDSTARTSIZE

        self.SetScrollRate(1,1)

        self.Bind(wx.EVT_PAINT, self.OnPaint)
        self.Bind(wx.EVT_ERASE_BACKGROUND, lambda x:None)
        self.Bind(wx.EVT_MOUSE_EVENTS, self.OnMouse)
        self.Bind(wx.EVT_CHAR,self.OnChar)
        
        # vars for handling mouse clicks
        self.dragid = -1
        self.lastpos = (0,0)
        self.originalpos = (0,0)
        
        self.dragStart   = None
        self.linkStart   = None
        self.linkEnd     = None
        self.posSelected = None
        self.negSelected = None
        self.lastCurrent = None
        
        self.SetRandomSeed(RAND_SEED)

        self.onodegrid = ONodeGrid(GRIDSTARTSIZE,GRIDSTARTSIZE)
        self.InitializeDrawing(self.pdc)
        #self.ScrollToCoords((0,0))

    def DoLoad(self, bAssembly=True, bSystem=False):
        typeString = "Oritatami assembly files (*.oass)|*.oass"
        if bAssembly == False:
            typeString = "Oritatami system files (*.orit)|*.orit"
        elif bAssembly == True and bSystem == True:
            typeString = "Oritatami system/seed files (*.orit;*.oass)|*.orit;*.oass"
        openFileDialog = wx.FileDialog(self, "Open", "", "", 
                                       typeString,
                                       wx.FD_OPEN | wx.FD_FILE_MUST_EXIST)
        openFileDialog.ShowModal()
        fileName = openFileDialog.GetPath()
        openFileDialog.Destroy()
        if (len(fileName) > 0):
            
            dlg = wx.TextEntryDialog(self, "You can set your random seed below\n(use -1 to choose a new random value):", caption="Set random number seed", defaultValue=str(self.randSeed))
            if dlg.ShowModal() == wx.ID_OK:
                try:
                    self.randSeed = int(dlg.GetValue())
                except ValueError:
                    globalApp.ShowErrorDialog("Random seed value must be an integer.")
                    dlg.Destroy()
                    return
            dlg.Destroy()

            self.SetRandomSeed(self.randSeed)

            self.onodegrid = ONodeGrid(GRIDSTARTSIZE,GRIDSTARTSIZE)
            if fileName.endswith(".orit"):
                self.onodegrid.LoadOritatamiSystem(fileName)
            elif fileName.endswith(".oass"):
                self.onodegrid.LoadOritatamiAssembly(fileName, True)

                #if len(self.onodegrid.ONodeTypes) > 0:
                    # Randomly set the current drawing bead type to one of the bead types
                    #self.drawingModeBeadType = self.onodegrid.ONodeTypes[self.onodegrid.ONodeTypes.keys()[0]]
                #else:
            self.drawingModeBeadType = None
            self.InitializeDrawing(self.pdc)
            self.app.AddBeadTypeWindow()
            self.app.DrawBeadCycle()
            self.Refresh()
            self.ScrollToCoords((0,0))

    def DoSaveSystemDef(self, bSaveAs=False):
        fileName = globalApp.systemDefinitionFileFullPath
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save System Definition As", "", "", 
                                           "Oritatami system files (*.orit)|*.orit",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                globalApp.ShowErrorDialog("You must provide a file name to save the oritatami system definition to.")
                return False
            fileName = saveFileDialog.GetPath()
            globalApp.systemDefinitionFile = fileName
            saveFileDialog.Destroy()

        globalApp.systemDefinitionFileFullPath = fileName
        if os.path.isabs(fileName) == False:
            fileName = "./" + fileName
            fileName = os.path.abspath(fileName)

        try:
            btStr = "["
            for bt in self.onodegrid.ONodeTypes:
                if len(btStr) > 1:
                    btStr = btStr + ", "
                btStr = btStr + self.onodegrid.ONodeTypes[bt].Output()
            btStr = btStr + "]"
            btLine = "BeadTypes=" + btStr
            print("btLine: " + btLine)
            bcLine = "BeadCycle=" + str(self.onodegrid.beadCycle)
            print("bcLine: " + bcLine)
            dlLine = "DelayLength=" + str(self.onodegrid.delayLength)
            print("dlLine: " + dlLine)
            arLine = "Arity=" + str(self.onodegrid.arity)
            print("arLine:" + arLine)
            lwbLine = "LockWithoutBinding=" + str(self.onodegrid.lockWithoutBinding)
            print("lwbLine: " + lwbLine)
            bffsLine = "BindFullFloppySegment=" + str(self.onodegrid.bindFullFloppySegment)
            print("bffsLine:" + bffsLine)

            wFile = open(fileName, "w")
            wFile.write(btLine + "\n")
            wFile.write(bcLine + "\n")
            wFile.write(dlLine + "\n")
            wFile.write(arLine + "\n")
            wFile.write(lwbLine + "\n")
            wFile.write(bffsLine + "\n")
            wFile.close()
            globalApp.SetOutputMessage("Saved oritatami system definition to file '" + fileName + "'")
            return True
        except:
            globalApp.ShowErrorDialog("\nERROR: Unable to write oritatami system definition file: '" + fileName + "'\n")
            return False

    def DoSetSeed(self):
        dlg = wx.MessageDialog(self, "Are you sure you want to set the current assembly as the seed assembly?", caption="Confirm seed setting")
        if dlg.ShowModal() == wx.ID_OK:
            dlg.Destroy()
            assemblyList = self.GetAssemblyList()
            if len(assemblyList) == 0:
                globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
                return False
            self.onodegrid.seed = assemblyList
            self.onodegrid.lastSeedLoc = self.onodegrid.lastBeadLoc
        else:
            dlg.Destroy()

    #  Format for assembly file:
    #  [[<bead location>], [<bead location>], ...]
    #  Format for bead location:
    #  ["<type name>", (<coords>), (<parent coords>), [<list of coords bonded to>]]
    def DoSaveAssembly(self, bOnlySeed=False, bSaveAs=False):
        aStr = "assembly"
        AStr = "Assembly"
        fileName = globalApp.assemblyDefinitionFile
        if bOnlySeed == True:
            aStr = "seed"
            AStr = "Seed"
            fileName = globalApp.seedDefinitionFile
        if fileName is None or bSaveAs == True:
            saveFileDialog = wx.FileDialog(self, "Save " + AStr + " As", "", "", 
                                           "Oritatami " + aStr + " files (*.oass)|*.oass",
                                           wx.FD_SAVE | wx.FD_OVERWRITE_PROMPT)
            if not saveFileDialog.ShowModal() == wx.ID_OK:
                return False
            fileName = saveFileDialog.GetPath()
            saveFileDialog.Destroy()

        if globalApp.systemDefinitionFile is None or len(globalApp.systemDefinitionFile) < 1:
            globalApp.ShowErrorDialog("\nERROR: In order to save an assembly, you must first have a valid oritatami system definition file associated with it. Please save your system definition first.\n")
            return False

        systemDefString = "OritatamiSystemDef=" + globalApp.systemDefinitionFile
        assemblyString = "Assembly="
        if bOnlySeed == True:
            lastBeadString = "LastBead=" + str(self.onodegrid.lastSeedLoc)
        else:
            lastBeadString = "LastBead=" + str(self.onodegrid.lastBeadLoc)
        assemblyList = self.GetAssemblyList(bOnlySeed)
        if len(assemblyList) == 0:
            globalApp.ShowErrorDialog("The seed assembly is empty.  It must contain at least one bead.")
            return False
        assemblyString = assemblyString + str(assemblyList)
        
        origFileName = fileName
        if not fileName is None:
            if len(fileName) < 1:
                return False
            elif os.path.isabs(fileName) == False:
                fileName = "./" + fileName
                fileName = os.path.abspath(fileName)

            try:
                wFile = open(fileName, "w")
                wFile.write(systemDefString + "\n")
                wFile.write(assemblyString + "\n")
                wFile.write(lastBeadString)
                wFile.close()
                if bOnlySeed == True:
                    globalApp.seedDefinitionFile = origFileName
                else:
                    globalApp.assemblyDefinitionFile = origFileName
                globalApp.SetOutputMessage("Saved oritatami " + aStr + " to file '" + fileName + "'")
                return True
            except:
                globalApp.ShowErrorDialog("\nERROR: Unable to write oritatami " + aStr + " to file: '" + fileName + "'\n")
                return False

    def GetAssemblyList(self, bOnlySeed=False):
        assemblyList = []
        if bOnlySeed == True:
            assemblyList = self.onodegrid.seed
        else:
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                    locList = self.GetLocOutputList((i,j))
                    if not locList is None:
                        assemblyList.append(locList)
        return assemblyList

    def GetLocOutputList(self, loc):
        onode = self.onodegrid.GetLoc(loc)
        if onode is not None:
            #self.onodegrid.ONodeTypes[onode.type]
            locList = []
            locList.append(onode.type)
            locList.append(loc)
            if onode.parent is None:
                locList.append(None)
            else:
                locList.append(onode.parent.coords)
            locList.append(onode._bonds)
            return locList
        else:
            return None
        
    def DoSaveAll(self):
        if self.DoSaveSystemDef(bSaveAs=False) == True:
            if self.DoSaveAssembly(bOnlySeed=True, bSaveAs=False) == True:
                return True
        return False

    def SetRandomSeed(self, seed):
        self.randSeed = seed
        if self.randSeed < 0:
            self.randSeed = random.randint(0, sys.maxint)
        print("Seeding random number generator with " + str(self.randSeed))
        random.seed(self.randSeed)

    def DoNew(self):
        self.onodegrid = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)

        self.SetRandomSeed(self.randSeed)
        globalApp.seedDefinitionFile = None
        globalApp.assemblyDefinitionFile = None
        globalApp.systemDefinitionFile = None
        globalApp.systemDefinitionFileFullPath = None
        
        self.drawingModeBeadType = None
        self.InitializeDrawing(self.pdc)
        self.app.AddBeadTypeWindow()
        self.app.DrawBeadCycle()
        self.Refresh()
        self.ScrollToCoords((0,0))

        return False

    def Reset(self):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reset your system to the\nseed assembly and abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", defaultValue=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        try:
            result = int(dlg.GetValue())
            self.SetRandomSeed(result)
        except ValueError:
            globalApp.ShowErrorDialog("Random seed value must be an integer.")
            dlg.Destroy()
            return
        dlg.Destroy()
        self.onodegrid.ResetToSeed()
        self.SetRandomSeed(self.randSeed)
        self.InitializeDrawing(self.pdc)
        self.app.DrawBeadCycle()
        self.Refresh()

    def Reload(self):
        dlg = wx.TextEntryDialog(self, "Are you sure you want to reload your system\nand abandon the current assembly?\nIf so, you can set your random seed below\n(use -1 to choose a new random value):", defaultValue=str(self.randSeed))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        try:
            result = int(dlg.GetValue())
            self.SetRandomSeed(result)
        except ValueError:
            globalApp.ShowErrorDialog("Random seed value must be an integer.")
            dlg.Destroy()
            return
        dlg.Destroy()

        self.onodegrid = ONodeGrid(GRIDSTARTSIZE, GRIDSTARTSIZE)
        if not globalApp.seedDefinitionFile is None:
            self.onodegrid.LoadOritatamiAssembly(globalApp.seedDefinitionFile, True)
        elif not globalApp.systemDefinitionFile is None:
            self.onodegrid.LoadOritatamiSystem(globalApp.systemDefinitionFile, globalApp.systemDefinitionFileFullPath)  
        self.InitializeDrawing(self.pdc)
        self.app.DrawBeadCycle()
        self.Refresh()

    def SetBreakpoints(self):
        outStr = ""
        if self.breakOnNondeterminism == True:
            outStr = "NONDET"
            if len(self.breakpoints) > 0:
                outStr = outStr + ","
        for i in range(len(self.breakpoints)):
            outStr = outStr + str(self.breakpoints[i])
            if i <  (len(self.breakpoints)-1):
                outStr = outStr + ","
        dlg = wx.TextEntryDialog(self, "Please enter a comma separated list of numbers to break\nwhen those numbers of beads are contained in the assembly.\nUse 'NONDET' to break when nondeterminism is detected.", caption="Set Breakpoints", defaultValue=str(outStr))
        if not dlg.ShowModal() == wx.ID_OK:
            return
        result = dlg.GetValue()
        dlg.Destroy()
        rawBPs = [x.strip() for x in result.split(',')]
        self.breakOnNondeterminism = False
        self.breakpoints = []
        self.currentBreakpoint = -1
        for bp in rawBPs:
            if len(bp) > 0:
                if bp == "NONDET":
                    self.breakOnNondeterminism = True
                else:
                    try:
                        newBP = int(bp)
                        self.breakpoints.append(newBP)
                    except ValueError:
                        dlg = wx.MessageDialog(self, "Invalid breakpoints entered", caption="Error", style=wx.ICON_ERROR|wx.OK)
                        dlg.ShowModal()
                        dlg.Destroy()
                        self.breakpoints = []
                        self.breakOnNondeterminism = False 
                        return
        if len(self.breakpoints) > 0:
            self.breakpoints.sort()
            self.currentBreakpoint = 0

    def DoGo(self):
        if self.keepGoing == True:
            self.app.SetOutputMessage("ERROR: Cannot create another new thread, there is already one running!")
            return
        self.keepGoing = True
        try:
            self.goThread = thread.start_new_thread(self.GoForward, ())
        except:
            self.app.SetOutputMessage("ERROR: Failed to create thread!")
    
    def DoQuit(self):
        self.keepGoing = False
        self.showMouseOver = True
        #self.goThread.exit()                

    def GoForward(self):
        while self.keepGoing == True:
            self.showMouseOver = False
            if self.DoForwardStep() == False:
                self.DoQuit()
        print("Exiting thread")
        return
        #self.app.SetOutputMessage("Stopping fast forward simulation")

    def DoForwardStep(self):
        oldgridminx = self.onodegrid.minX
        usp = self.CalcUnscrolledPosition(0,0)
        sz = self.GetSize()
        center = (usp[0] + sz[0]/2, usp[1] + sz[1]/2)
        coords = self.GetCoordsFromPixels(center[0], center[1], True)

        (addedStat, addedList) = self.onodegrid.StepForward()
        if (addedStat == ORI_STAT_NO_SPACE) or (addedStat == ORI_STAT_MAX_FLOPPY) or (addedStat == ORI_STAT_ERROR):
            return False
        else:
            self._lock.acquire()
            self.DrawQueue = self.DrawQueue + addedList
            self._lock.release()
        if not self.onodegrid.minX == oldgridminx:
            #MJP: not quite right...
            self.InitializeDrawing(self.pdc)
            self.ScrollToCoords(coords)
            
        #self._lock.acquire()
        self.app.DrawBeadCycle()
        #self._lock.release()
        
        #self.app.SetOutputMessage("Segment to add: " + str(self.onodegrid.floppySegment))
        if self.app.showDisplayUpdates == True:
            self.DrawUpdate(self.pdc)
            self.Refresh()
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)

        if addedStat == ORI_STAT_NONDET and globalApp.canvas.breakOnNondeterminism == True:
            self.app.SetOutputMessage("Breakpoint encountered due to nondeterminism")
            return False
        elif (self.currentBreakpoint > -1) and (self.onodegrid.GetNumBeads() >= self.breakpoints[self.currentBreakpoint]):
            self.app.SetOutputMessage("Breakpoint encountered: " + str(self.breakpoints[self.currentBreakpoint]) + " beads met or exceeded for the first time")
            self.currentBreakpoint = self.currentBreakpoint + 1
            if self.currentBreakpoint >= len(self.breakpoints):
                self.currentBreakpoint = -1
            while (self.currentBreakpoint > -1) and (self.breakpoints[self.currentBreakpoint] < self.onodegrid.GetNumBeads()):
                self.currentBreakpoint = self.currentBreakpoint + 1
                if self.currentBreakpoint >= len(self.breakpoints):
                    self.currentBreakpoint = -1
            return False
        return True
    
    def ZoomThenRecenter(self, zoomIn=True):
            usp = self.CalcUnscrolledPosition(0,0)
            sz = self.GetSize()
            center = (usp[0] + sz[0]/2, usp[1] + sz[1]/2)
            coords = self.GetCoordsFromPixels(center[0], center[1], True)
            self.DoScale(zoomIn)
            self.ScrollToCoords(coords)
    
    def ConvertEventCoords(self, event):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        return (event.GetX() + (xView * xDelta),
            event.GetY() + (yView * yDelta))

    def OffsetRect(self, r):
        xView, yView = self.GetViewStart()
        xDelta, yDelta = self.GetScrollPixelsPerUnit()
        r.OffsetXY(-(xView*xDelta),-(yView*yDelta))

    def OnChar(self, evt):
        keycode = evt.GetKeyCode()
        if keycode == ord("a") or keycode == ord("A"):
            self.AddTileCopiesToDrawing(self.pdc)
            self.Refresh()
        elif keycode == ord("f") or keycode == ord("F"):
            self.DoForwardStep()
        elif keycode == ord("p") or keycode == ord("P"):
            self.PrintDebugInfo()
        elif keycode == ord("r") or keycode == ord("R"):
            self.Reset()
        elif keycode == ord("s") or keycode == ord("S"):
            self.DoSaveAll()
        elif keycode == ord("h") or keycode == ord("H"):
            self.ShowHelp()
        elif keycode == ord("l") or keycode == ord("L"):
            self.DoLoad(bAssembly=True, bSystem=True)
        elif keycode == ord("=") or keycode == ord("+"):
            self.DoScale(True)
        elif keycode == ord("-") or keycode == ord("_"):
            self.DoScale(False)
        elif keycode == ord("c") or keycode == ord("C"):
            self.ScrollToCoords((0,0))
        elif keycode == ord("t") or keycode == ord("T"):
            self.ToggleGrid()
        elif keycode == ord("g") or keycode == ord("G"):
            self.DoGo()
        elif keycode == ord("q") or keycode == ord("Q"):
            self.DoQuit()
        elif keycode == ord("m") or keycode == ord("M"):
            self.MessageToggle(not self.app.showMessages)
            self.app.menuView.Check(self.app.itemTM.GetId(), self.app.showMessages)
        elif keycode == ord("d") or keycode == ord("D"):
            self.DisplayToggle(not self.app.showDisplayUpdates)
            self.app.menuView.Check(self.app.itemTD.GetId(), self.app.showDisplayUpdates)
        elif keycode == ord("b") or keycode == ord("B"):
            self.SetBreakpoints()
        #elif keycode == wx.WXK_DELETE or keycode == wx.WXK_BACK:
            #self.DoDelete()

    def ShowHelp(self):
        self.app.OnHelp(None)

    def PrintDebugInfo(self):        
        #self.app.SetOutputMessage("some stuff")
        return

    def DoScale(self, bIncrease):
        #wx.MessageBox("Oops, this doesn't really work yet...", "Zooming not supported", wx.ICON_EXCLAMATION)
        #return
        if bIncrease == True:
            self.scale = self.scale * 2.0
        else:
            self.scale = self.scale / 2.0
        print("Scale set to " + str(self.scale))
        
        self.SetVirtualSize((self.winWidth*self.scale, self.winHeight*self.scale))

        self.Refresh()


    def OnMouse(self, event):
        
        #=======================================================================
        # if event.IsPageScroll():
        #     print("is page scroll")
        # if event.MiddleIsDown():
        #     print("middle is down")
        #     if event.Dragging():
        #         print("Dragging")
        #=======================================================================
            
        x,y = self.ConvertEventCoords(event)
        
        if self.showMouseOver == True:
            x2,y2 = self.GetCoordsFromPixels(x, y)
    
            if x2 == None:
                self.frame.SetStatusText("", 0)
                self.frame.SetStatusText("", 1)
            else:
                self.frame.SetStatusText("Coords: (" + str(int(x2)) + "," + str(int(y2)) + ")", 0)
                onode = self.onodegrid.GetLoc((x2,y2))
                if not onode is None:
                    outStr = "Bead type: " + str(onode.type) + " (Binds to " + str(self.onodegrid.ONodeTypes[onode.type].bondToTypes) + ")"
                    self.frame.SetStatusText(outStr, 1)
                else:
                    self.frame.SetStatusText("", 1)
        
        if event.LeftDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                self.SetCursor(wx.CursorFromImage(self.grabbingCursorImg))
                self.dragStart = self.ScreenToClient(wx.GetMousePosition())
                self.dragStartScrollPos = (self.GetScrollPos(wx.HORIZONTAL), self.GetScrollPos(wx.VERTICAL))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if self.drawingModeBeadType is None:
                    globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                else:
                    x2,y2 = self.GetCoordsFromPixels(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.onodegrid.GetLoc((x2, y2)) is None:
                            parentLoc = None
                            if (self.onodegrid.lastBeadLoc is not None) and self.onodegrid.AreNbrLocs((x2, y2), self.onodegrid.lastBeadLoc) == True:
                                parentLoc = self.onodegrid.lastBeadLoc
                            self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.negSelected is not None):
                    xs,ys = self.GetCoordsFromPixels(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.onodegrid.GetLoc(self.linkStart)
                        if self.negSelected is not None:
                            saved = self.negSelected.coords
                            self.negSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.posSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.GetCoordsFromPixels(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.onodegrid.GetLoc(self.linkStart)
                        self.posSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.onodegrid.GetLoc(self.linkStart)
                            childNode = self.onodegrid.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (self.onodegrid.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        if parentNode.child is not None:
                                            globalApp.ShowErrorDialog("The node at " + str(self.linkStart) + " already has a child")
                                        elif childNode.parent is not None:
                                            globalApp.ShowErrorDialog("The node at " + str(self.linkEnd) + " already has a parent")
                                        else:
                                            bAlreadyBonded = False
                                            for n in range(parentNode.NumBonds()):
                                                b = parentNode.GetNthBond(n)
                                                if b == self.linkEnd:
                                                    bAlreadyBonded = True
                                            if bAlreadyBonded == True:
                                                globalApp.ShowErrorDialog("Nodes must not already be bonded to be linked as parent and child")
                                            else:
                                                parentNode.child = childNode
                                                childNode.parent = parentNode
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        globalApp.ShowErrorDialog("Nodes must be neighbors to be linked")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    if (self.onodegrid.AreNbrLocs(self.linkStart, self.linkEnd) == True):
                                        bAlreadyBonded = False
                                        for n in range(parentNode.NumBonds()):
                                            b = parentNode.GetNthBond(n)
                                            if b == self.linkEnd:
                                                bAlreadyBonded = True
                                        if bAlreadyBonded == True:
                                            globalApp.ShowErrorDialog("Nodes are already bonded")
                                        elif (parentNode.child == childNode) or (childNode.child == parentNode):
                                            globalApp.ShowErrorDialog("Nodes cannot be bonded if they are linked as parent and child")
                                        else:
                                            bDoIt = True
                                            numP = parentNode.NumBonds()
                                            if (self.confirmOverArity == True) and (numP >= self.onodegrid.arity):
                                                msg = "The bead at location " + str(parentNode.coords) + " currently has " + str(numP)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.onodegrid.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == ORIT_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == ORIT_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            numC = childNode.NumBonds()
                                            if (bDoIt == True) and (self.confirmOverArity == True) and (numC >= self.onodegrid.arity):
                                                msg = "The bead at location " + str(childNode.coords) + " currently has " + str(numC)
                                                if numP == 1:
                                                    msg = msg + " bond"
                                                else:
                                                    msg = msg + " bonds"
                                                msg = msg + ", and the arity parameter is set to " + str(self.onodegrid.arity) + ".\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm exceeding arity", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == ORIT_ID_OK_NOCHECK:
                                                    self.confirmOverArity = True
                                                elif retId == ORIT_ID_OK_CHECK:
                                                    self.confirmOverArity = False
                                                else:
                                                    bDoIt = False
                                            if (bDoIt == True) and (self.confirmInvalidBonds == True) and (self.onodegrid.ONodeTypes[parentNode.type].BondsToNodeType(childNode.type) == False):
                                                msg = "Bead types '" + str(parentNode.type) + "' and '" + str(childNode.type) + "' don't form bonds."
                                                msg = msg + "\nDo you want to create the new bond anyway?" 
                                                dlg = DontShowAgainDialog(self, wx.ID_ANY, "Confirm bond between non-bonding types", msg)
                                                dlg.CenterOnScreen()
                                                retId = dlg.ShowModal()
                                                dlg.Destroy()
                                                if retId == ORIT_ID_OK_NOCHECK:
                                                    self.confirmInvalidBonds = True
                                                elif retId == ORIT_ID_OK_CHECK:
                                                    self.confirmInvalidBonds = False
                                                else:
                                                    bDoIt = False
                                            if bDoIt == True:
                                                self.onodegrid.AddBond(self.linkStart, self.linkEnd)
                                                self.posSelected = None
                                                self.ManuallyDrawNodes([self.linkStart, self.linkEnd])
                                                self.linkStart = None
                                                self.linkEnd = None
                                    else:
                                        globalApp.ShowErrorDialog("Nodes must be neighbors to be bonded")

        elif event.RightDown():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                #self.SetCursor(wx.CursorFromImage(self.grabbingCursorImg))
                doScale = True
                if (event.AltDown() == True) or (event.ControlDown() == True) or (event.ShiftDown() == True):
                    doScale = False
                x2,y2 = self.GetCoordsFromPixels(x, y, True)
                self.DoScale(doScale)
                self.ScrollToCoords((x2,y2))
                sz = self.GetSize()
                self.WarpPointer(sz[0]/2, sz[1]/2)
                #self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))

            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                x2,y2 = self.GetCoordsFromPixels(x, y)
                if (x2 is not None) and (y2 is not None):
                    if self.onodegrid.GetLoc((x2, y2)) is not None:
                        self.ManuallyRemoveNode((x2, y2))
                        
            elif (self.mouseMode == ORIT_MODE_DRAW_LINK) or (self.mouseMode == ORIT_MODE_DRAW_BOND):
                if (self.linkStart is None) or (self.posSelected is not None):
                    xs,ys = self.GetCoordsFromPixels(x, y)
                    if (xs is None) or (ys is None):
                        self.linkStart = None
                    else:
                        self.linkStart = (xs,ys)
                        node = self.onodegrid.GetLoc(self.linkStart)
                        if self.posSelected is not None:
                            saved = self.posSelected.coords
                            self.posSelected = None
                            self.ManuallyDrawNodes([saved])
                        self.negSelected = node
                        self.ManuallyDrawNodes([self.linkStart])
                else:
                    xe,ye = self.GetCoordsFromPixels(x, y)
                    if self.linkStart[0] == xe and self.linkStart[1] == ye:
                        node = self.onodegrid.GetLoc(self.linkStart)
                        self.negSelected = None
                        self.ManuallyDrawNodes([self.linkStart])
                        self.linkStart = None
                    else:
                        if (xe is None) or (ye is None):
                            self.linkEnd = None
                        else:
                            self.linkEnd = (xe,ye)
                        if (self.linkStart is not None) and (self.linkEnd is not None):
                            parentNode = self.onodegrid.GetLoc(self.linkStart)
                            childNode = self.onodegrid.GetLoc(self.linkEnd)
                            if (parentNode is not None) and (childNode is not None):
                                bDidIt = False
                                if self.mouseMode == ORIT_MODE_DRAW_LINK:
                                    if (parentNode.child == childNode):
                                        parentNode.child = None
                                        childNode.parent = None
                                        bDidIt = True
                                    elif (childNode.child == parentNode):
                                        childNode.child = None
                                        parentNode.parent = None
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.InitializeDrawing(self.pdc)
                                    else:
                                        globalApp.ShowErrorDialog("Cannot remove link between beads which aren't linked.")
                                elif self.mouseMode == ORIT_MODE_DRAW_BOND:
                                    bAlreadyBonded = False
                                    for n in range(parentNode.NumBonds()):
                                        b = parentNode.GetNthBond(n)
                                        if b == self.linkEnd:
                                            bAlreadyBonded = True
                                    if bAlreadyBonded == True:
                                        self.onodegrid.RemoveBond(self.linkStart, self.linkEnd)
                                        bDidIt = True
                                    if bDidIt == True:
                                        self.negSelected = None
                                        self.linkStart = None
                                        self.linkEnd = None
                                        self.InitializeDrawing(self.pdc)
                                    else:
                                        globalApp.ShowErrorDialog("Cannot remove bond between beads which don't have a bond.")

        if event.LeftUp():
            self.dragStart = None
            self.SetCursor(wx.StockCursor(wx.CURSOR_ARROW))

        elif event.Dragging():
            if self.mouseMode == ORIT_MODE_PAN_ZOOM:
                if self.dragStart is not None:
                    #self.SetScrollRate(1,1)
                    new = self.ScreenToClient(wx.GetMousePosition())
                    dx = int((self.dragStart.x-new.x)/1)
                    dy = int((self.dragStart.y-new.y)/1)
                    spx = self.GetScrollPos(wx.HORIZONTAL)
                    spy = self.GetScrollPos(wx.VERTICAL)
                    self.Scroll(spx+dx, spy+dy)
                    self.dragStart = new
            
            elif self.mouseMode == ORIT_MODE_DRAW_BEAD:
                if event.LeftIsDown():
                    if self.drawingModeBeadType is None:
                        globalApp.ShowErrorDialog("A bead type must be selected in the bead type editor window")
                    else:
                        x2,y2 = self.GetCoordsFromPixels(x, y)
                        if (x2 is not None) and (y2 is not None):
                            if self.onodegrid.GetLoc((x2, y2)) is None:
                                parentLoc = None
                                if (self.onodegrid.lastBeadLoc is not None) and self.onodegrid.AreNbrLocs((x2, y2), self.onodegrid.lastBeadLoc) == True:
                                    parentLoc = self.onodegrid.lastBeadLoc
                                self.ManuallyAddNode(self.drawingModeBeadType.name, (x2, y2), parentLoc)
                elif event.RightIsDown():
                    x2,y2 = self.GetCoordsFromPixels(x, y)
                    if (x2 is not None) and (y2 is not None):
                        if self.onodegrid.GetLoc((x2, y2)) is not None:
                            self.ManuallyRemoveNode((x2, y2))

    def ManuallyAddNode(self, typeName, coords, parentCoords):
        self.onodegrid.AddNode(self.drawingModeBeadType.name, coords, parentCoords)
        self.ManuallyDrawNodes([coords])
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)
        
    def ManuallyRemoveNode(self, coords):
        self.onodegrid.RemoveNode(coords)
        self.InitializeDrawing(self.pdc)
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)
        
    def ManuallyDrawNodes(self, coordList):
        self._lock.acquire()
        for coords in coordList:
            self.DrawQueue.append(coords)
        self._lock.release()
        self.DrawUpdate(self.pdc)
        self.Refresh()

    def ScrollToCoords(self, coords):
        if (not coords == None) and (not coords[0] == None) and (not coords[1] == None):
            (x,y) = self.GetPixelsForCoords(coords[0], coords[1])
            x = x * self.scale
            y = y * self.scale
            (xs, ys) = self.GetScrollPixelsPerUnit()
            (xsz,ysz) = self.GetSizeTuple()
            self.Scroll((x-xsz/2+self.sideLen*self.scale)/xs, (y-ysz/2+self.sideLen*self.scale)/ys)


    def ColoredPen(self, color, width = 2):
        if not self.pen_cache.has_key( (color, width) ):
            self.pen_cache[(color, width)] = wx.Pen(color, width)
        return self.pen_cache[(color, width)]

    def ColoredBrush(self, colorName):
        if not self.brush_cache.has_key(colorName):
            if type(colorName) is str:
                if "(" in colorName:
                    colorTuple = eval(colorName)
                    if len(colorTuple) > 3:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2], colorTuple[3])
                    else:
                        color = wx.Colour(colorTuple[0], colorTuple[1], colorTuple[2])
                else:
                    color = wx.TheColourDatabase.Find(colorName)
                self.brush_cache[colorName] = wx.Brush(color)
        return self.brush_cache[colorName]


    def OnPaint(self, event):
        # Create a buffered paint DC.  It will create the real
        # wx.PaintDC and then blit the bitmap to it when dc is
        # deleted.  
        dc = wx.BufferedPaintDC(self)
        # use PrepateDC to set position correctly
        self.PrepareDC(dc)
        # we need to clear the dc BEFORE calling PrepareDC
        bg = wx.Brush(self.GetBackgroundColour())
        dc.SetBackground(bg)
        dc.Clear()

        dc.SetUserScale(self.scale, self.scale)

        # create a clipping rect from our position and size
        # and the Update Region
        xv, yv = self.GetViewStart()
        dx, dy = self.GetScrollPixelsPerUnit()
        x, y   = (xv * dx, yv * dy)
        rgn = self.GetUpdateRegion()
        rgn.Offset(x,y)
        r = rgn.GetBox()
        # draw to the dc using the calculated clipping rect
        self.pdc.DrawToDCClipped(dc,r)



    def DrawUpdate(self, dc):
       
        dc.BeginDrawing()
        
        # Thread-safely make a copy of the list to draw then clear it out
        self._lock.acquire()
        copiedList = list(self.DrawQueue)
        if self.lastCurrent is not None:    # Redraw the bead that was current at the last step
            copiedList.append(self.lastCurrent)
        self.DrawQueue = []
        self._lock.release()
        
        for e in copiedList:
            onode = self.onodegrid.GetLoc((e[0], e[1]))
            if onode is not None:
                self.DrawAtCoords(dc, e[0], e[1], onode, self.onodegrid)

        #=======================================================================
        # for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
        #     for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
        #         onode = self.onodegrid.GetLoc((i,j))
        #         if onode is not None:
        #             self.DrawAtCoords(dc, i, j, onode, self.onodegrid)
        #=======================================================================

        if len(self.RemoveQueue) > 0:
            self._lock.acquire()
            self.InitializeDrawing(self.pdc)
            self.RemoveQueue = []
            self._lock.release()
            
        for e in self.RemoveQueue:
            dc.RemoveId(e)
            #pen = self.ColoredPen("BLACK")
            #dc.SetPen(pen)
            #dc.SetBrush(self.ColoredBrush("WHITE"))
            #dc.DrawRectangle(0,0,300,300)                            
        self.RemoveQueue = []
        dc.EndDrawing()
        
    def InitializeDrawing(self, dc):
       
        # Compute new dimensions
        self.gridWidth = self.onodegrid.GetWidth()
        self.gridHeight = self.onodegrid.GetHeight()
        
        self.winWidth  = self.sideLen * self.gridWidth  + self.GetHorizOffset(self.gridWidth)
        self.winHeight = self.GetVertOffset(self.gridHeight)

        self.SetVirtualSize((self.winWidth*self.scale, self.winHeight*self.scale))

        #  Remove old data
        dc.RemoveAll()
        dc.Clear()
        self.Refresh()

        dc.BeginDrawing()
        dc.SetPen(self.gridPen)
        
        if (self.showGrid == SHOW_GRID_AND_COORDS) or (self.showGrid == SHOW_GRID_ONLY):
            # Draw horizontal lines
            width = self.sideLen * (self.gridWidth - 1)
            for i in range(self.gridHeight):
                hOffset = self.GetHorizOffset(i)
                vOffset = self.GetVertOffset(i)
                dc.DrawLine(hOffset, vOffset, hOffset + width, vOffset)
            
            # Draw negative sloped vertical lines
            height = self.GetVertOffset(self.gridHeight - 1)
            bottomHOffset = self.GetHorizOffset(self.gridHeight - 1)
            for i in range(self.gridWidth):
                dc.DrawLine(i * self.sideLen, 0, bottomHOffset + (i * self.sideLen), height)
                
            # Draw positive sloped vertical lines
            #MJP: assuming a square grid (i.e. same number vertical and horizontal cells)
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                startPt = self.GetPixelsForCoords(i, self.onodegrid.minY)
                endPt = self.GetPixelsForCoords(self.onodegrid.minX,i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])
                startPt = self.GetPixelsForCoords(i, self.onodegrid.maxY)
                endPt = self.GetPixelsForCoords(self.onodegrid.maxX, i)
                dc.DrawLine(startPt[0], startPt[1], endPt[0], endPt[1])

        if self.showGrid == SHOW_GRID_AND_COORDS:
            for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
                for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                    (xp,yp) = self.GetPixelsForCoords(i,j)
                    dc.DrawText(str(i) + "," + str(j), xp+8, yp-2)
        
        for i in range(self.onodegrid.minX, self.onodegrid.maxX+1):
            for j in range(self.onodegrid.minY, self.onodegrid.maxY+1):
                onode = self.onodegrid.GetLoc((i,j))
                if onode is not None:
                    self.DrawAtCoords(dc, i, j, onode, self.onodegrid)

        dc.EndDrawing()
        self.Refresh()
        self.frame.SetStatusText("Number of beads: " + str(self.onodegrid.GetNumBeads()), 2)


    def GetHorizOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(60))
    
    def GetVertOffset(self, y):
        return y * self.sideLen * math.cos(math.radians(30))

    def GetPixelsForCoords(self, xCoord, yCoord):
        xCoordAdj = xCoord - self.onodegrid.minX
        yCoordAdj = yCoord - self.onodegrid.minY
        xPixel = (xCoordAdj * self.sideLen) + self.GetHorizOffset(yCoordAdj)
        yPixel = self.GetVertOffset(yCoordAdj)
        return (xPixel,yPixel)

    def GetCoordsFromPixels(self, xPixel, yPixel, bGapOkay=False):
        xPixel = xPixel / self.scale
        yPixel = yPixel / self.scale
        yCoord = (yPixel / (self.sideLen * math.cos(math.radians(30)))) + self.onodegrid.minX
        xCoord = ((xPixel - self.GetHorizOffset(yCoord - self.onodegrid.minY)) / self.sideLen) + self.onodegrid.minX
        (x2,x1) = math.modf(xCoord)
        (y2,y1) = math.modf(yCoord)
        if ((abs(x2) < 0.2 or abs(x2) > 0.8) and (abs(y2) < 0.2 or abs(y2) > 0.8)) or (bGapOkay == True):
            retX = int(round(xCoord))
            retY = int(round(yCoord))
            if (retX >= self.onodegrid.minX and retX <= self.onodegrid.maxX and retY >= self.onodegrid.minY and retY <= self.onodegrid.maxY):
                return (retX, retY)
        return (None,None)

    def DrawAtCoords(self, dc, xCoord, yCoord, onode, grid):
        
        (xPixel,yPixel) = self.GetPixelsForCoords(xCoord, yCoord)

        if not onode.parent is None:
            (xPixelP,yPixelP) = self.GetPixelsForCoords(onode.parent.coords[0], onode.parent.coords[1])
            dc.SetPen(self.linkPen)
            
            # Draw parent-child bond
            dc.DrawLine(xPixel,yPixel,xPixelP,yPixelP)
            
            # Draw directional 'arrow' lines
            frac = 0.1
            midpoint = (int((xPixel+xPixelP)/2), int((yPixel+yPixelP)/2))
            slope = (xPixel-xPixelP, yPixel-yPixelP)
            midpoint = (midpoint[0]+slope[0]*frac, midpoint[1]+slope[1]*frac)
            invSlope = ((self.sideLen-abs(slope[0]))/self.sideLen, (self.sideLen-abs(slope[1]))/self.sideLen)

            linepoint = (int(midpoint[0]-slope[0]*frac), int(midpoint[1]-slope[1]*frac))

            mult = 10
            if (slope[0] == 0) or (slope[1] == 0):  # Horizontal
                pt1 = (int(linepoint[0]+invSlope[0]*7), int(linepoint[1]-invSlope[1]*7))
                pt2 = (int(linepoint[0]-invSlope[0]*7), int(linepoint[1]+invSlope[1]*7))
            elif ((slope[0] < 0) and (slope[1] < 0)) or ((slope[0] > 0) and (slope[1] > 0)):
                pt1 = (int(linepoint[0]+invSlope[0]*mult), int(linepoint[1]-invSlope[1]*mult))
                pt2 = (int(linepoint[0]-invSlope[0]*mult), int(linepoint[1]+invSlope[1]*mult))
            else:
                pt1 = (int(linepoint[0]+invSlope[0]*mult), int(linepoint[1]+invSlope[1]*mult))
                pt2 = (int(linepoint[0]-invSlope[0]*mult), int(linepoint[1]-invSlope[1]*mult))
            
            dc.DrawLine(midpoint[0], midpoint[1], pt1[0], pt1[1])
            dc.DrawLine(midpoint[0], midpoint[1], pt2[0], pt2[1])

        numBonds = onode.NumBonds()
        if numBonds > 0:
            dc.SetPen(self.bondPen)
            for n in range(numBonds):
                b = onode.GetNthBond(n)
                (x2, y2) = self.GetPixelsForCoords(onode.coords[0], onode.coords[1])
                (x3, y3) = self.GetPixelsForCoords(b[0], b[1])
                dc.DrawLine(x2, y2, x3, y3)
        dc.SetPen(self.gridPen)

        if onode.type in grid.ONodeTypes:
            nodeColor = grid.ONodeTypes[onode.type].color
        else:
            nodeColor = "Black"
        pointBrush = self.ColoredBrush(nodeColor)
        dc.SetBrush(pointBrush)
        dc.DrawCirclePoint((xPixel,yPixel), 10)

        if onode == self.posSelected:
            pointBrush = self.ColoredBrush("Green")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
        elif onode == self.negSelected:
            pointBrush = self.ColoredBrush("Red")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
        elif onode.coords == self.onodegrid.lastBeadLoc:
            pointBrush = self.ColoredBrush("White")
            dc.SetBrush(pointBrush)
            dc.DrawCirclePoint((xPixel,yPixel), 5)
            self.lastCurrent = onode.coords
        
    def ToggleGrid(self):
            self.showGrid = (self.showGrid + 1) % 3
            self.InitializeDrawing(self.pdc)
            
    def MessageToggle(self, bShow):
        self.app.showMessages = bShow
        if bShow == True:
            self.app.DrawBeadCycle()
            for msg in self.app.messageQueue:
                self.app.SetOutputMessage(msg)
        self.app.messageQueue = []

    def DisplayToggle(self, bShow):
        self.app.showDisplayUpdates = bShow
        if self.app.showDisplayUpdates == True:
            self.DrawUpdate(self.pdc)
            self.Refresh()

#---------------------------------------------------------------------------

class OritatamiBeadBox(wx.StaticBox):
    
    def __init__(self, beadType, panel, panelSizer, parentApp, beadBoxList):
        wx.StaticBox.__init__(self, panel, wx.ID_ANY, "", size=(200,100))
        self.parentPanel = panel
        self.parentPanelSizer = panelSizer
        self.parentApp = parentApp
        self.beadBoxList = beadBoxList
        self.beadType = beadType
        buttonLabel = "Add"
        if beadType is not None:
            buttonLabel = "Set"
        else:
            self.beadType = ONodeType("", "Black", [])
        #self.beadBox = wx.StaticBox(panel, wx.ID_ANY, "", size=(100,100))        
        self.beadBoxSizer = wx.StaticBoxSizer(self, wx.VERTICAL)

        self.beadBoxSizer.Add((0, 10), 0)  
        nameSizer = wx.BoxSizer(wx.HORIZONTAL)
        self.applyButton = wx.Button(self, wx.ID_ANY, buttonLabel, size=(35,-1))
        self.applyButton.Bind(wx.EVT_MOUSE_EVENTS, self.OnApply)
        nameSizer.Add(self.applyButton, 1)

        self.colorButton = SBitmapButton(self, wx.ID_ANY, self.GetBeadButtonBitmap(self.beadType.color))
        self.colorButton.Bind(wx.EVT_MOUSE_EVENTS, self.PickColor)  
        nameSizer.Add(self.colorButton, 1)
        nameSizer.AddSpacer(5)

        lblname = wx.StaticText(self, label="Name:")
        lblname.Bind(wx.EVT_MOUSE_EVENTS, self.parentApp.OnBeadPanelClickStatText)
        nameSizer.Add(lblname, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        nameSizer.AddSpacer(5)

        self.editName = wx.TextCtrl(self, size=(60, -1), value=self.beadType.name)
        nameSizer.Add(self.editName)
        
        if beadType is not None:
            self.deleteButton = wx.Button(self, wx.ID_ANY, "X", size=(2,30))
            self.deleteButton.Bind(wx.EVT_MOUSE_EVENTS, globalApp.OnBeadTypeDelete)
            nameSizer.Add(self.deleteButton, 1)

        self.beadBoxSizer.Add(nameSizer)
        self.beadBoxSizer.AddSpacer(5)
        
        bondsSizer = wx.BoxSizer(wx.HORIZONTAL)
        lblbonds = wx.StaticText(self, label="Bonds:")
        lblbonds.Bind(wx.EVT_MOUSE_EVENTS, self.parentApp.OnBeadPanelClickStatText)
        bondsSizer.Add(lblbonds, 0, wx.ALIGN_CENTER_VERTICAL, 5)
        bondsSizer.AddSpacer(5)

        bondTypeList = ""
        for bt in self.beadType.bondToTypes:
            if len(bondTypeList) > 0:
                bondTypeList = bondTypeList + ","
            bondTypeList = bondTypeList + bt
        self.editBonds = wx.TextCtrl(self, size=(155, -1), value=bondTypeList)
        bondsSizer.Add(self.editBonds)
        self.beadBoxSizer.Add(bondsSizer)

        self.beadBoxSizer.Add((0, 10), 0)
        panelSizer.Add(self.beadBoxSizer, 2)

    def OnApply(self, event):
        if event.LeftDown():
            box = event.EventObject.Parent
            bAdding = False
            if box.beadType.name == "":  # Creating a new bead type
                bAdding = True
            name = str(box.editName.GetValue())
            if len(name) == 0:
                globalApp.ShowErrorDialog("Bead type name cannot be blank")
                return
            elif (" " in name) or ("," in name) or (";" in name) :
                globalApp.ShowErrorDialog("Bead type name cannot contain a space, comma, or semicolon")
                return
            bondToTypes = str(box.editBonds.GetValue())
            if (" " in bondToTypes) or (";" in bondToTypes):
                globalApp.ShowErrorDialog("Bond types list should be a comma\nseparated list of bead type names")
                return
            typeList = []
            if len(bondToTypes) > 0:
                typeList = [str(t) for t in bondToTypes.split(",")]
            notFoundList = []
            for t in typeList:
                bFound = False
                for bb in self.beadBoxList:
                    if t == self.beadBoxList[bb].beadType.name:
                        bFound = True
                if bFound == False:
                    notFoundList.append(t)
            if len(notFoundList) > 0:
                globalApp.ShowWarningDialog("Warning: bond type list contains names of nonexistent bead types:\n" + str(notFoundList))

            if bAdding:
                
                for bb in self.beadBoxList:
                    if self.beadBoxList[bb].beadType.name == name:
                        globalApp.ShowErrorDialog("A bead type named '" + name + "' already exists.\nYou must choose a new name for a new type.")
                        return

                newBeadType = ONodeType(name, box.beadType.color, typeList)
                bb = OritatamiBeadBox(newBeadType, box.parentPanel, box.parentPanelSizer, box.parentApp, box.beadBoxList)
                box.beadBoxList[name] = bb
                box.parentPanelSizer.Layout()
                box.parentApp.frame._mgr.Update()
                
                # Reset the form
                box.beadType = ONodeType("", "Black", [])
                box.editName.SetValue("")
                box.editBonds.SetValue("")
                box.colorButton.SetBitmapLabel(self.GetBeadButtonBitmap(box.beadType.color))

                # Put the new bead type into simulator's list too
                box.parentApp.canvas.onodegrid.ONodeTypes[name] = newBeadType
            else:
                self.UpdateBeadType(box, box.beadType.name, name, box.beadType.color, typeList)
            
            if box.parentApp.canvas.onodegrid.FixBeadBondSymmetry(True, name, typeList) == True:
                for btName in box.parentApp.canvas.onodegrid.ONodeTypes:
                    bb = self.beadBoxList[btName]
                    btTypes = box.parentApp.canvas.onodegrid.ONodeTypes[btName].bondToTypes
                    btStr = ""
                    for bt in btTypes:
                        if len(btStr) > 0:
                            btStr = btStr + ","
                        btStr = btStr + bt
                    bb.editBonds.SetValue(btStr)
                    bb.editBonds.Refresh()

    
    def UpdateBeadType(self, beadBox, oldName, newName, newColor, newBondToTypes):
        #print("updating bead type " + oldName)
        #oldBeadType = beadBox.beadType

        if not oldName == newName:
            # Ask user if old bonds should be redone and/or if beads that used to bond to it should be updated to still bond to it
            #dlg = wx.MessageDialog(globalApp.frame, "Would you like bead types that bond to this type with its old name to bind\nto it with its new name and to update all beads in the current assembly?",
            #                                        "Update other bead types?", wx.YES_NO | wx.ICON_QUESTION)
            #result = (dlg.ShowModal() == wx.ID_YES)
            #dlg.Destroy()
            #if result == True:
            for nodeTypeName in beadBox.parentApp.canvas.onodegrid.ONodeTypes:
                nodeType = beadBox.parentApp.canvas.onodegrid.ONodeTypes[nodeTypeName]
                if nodeType.BondsToNodeType(oldName):
                    nodeType.bondToTypes.remove(oldName)
                    nodeType.bondToTypes.append(newName)
            for bb in self.beadBoxList:
                beadBox2 = self.beadBoxList[bb]
                nodeType = beadBox2.beadType
                if nodeType.BondsToNodeType(newName):
                    bondTypeList = ""
                    for bt in nodeType.bondToTypes:
                        if len(bondTypeList) > 0:
                            bondTypeList = bondTypeList + ","
                        bondTypeList = bondTypeList + bt
                    beadBox2.editBonds.SetValue(bondTypeList)
                    beadBox2.editBonds.Refresh()

            for index, item in enumerate(newBondToTypes):
                if item == oldName:
                    newBondToTypes[index] = newName

            beadBox.parentPanelSizer.Layout()
            beadBox.parentApp.frame._mgr.Update()

        newBeadType = ONodeType(newName, newColor, newBondToTypes)
        beadBox.beadType = newBeadType

        del beadBox.parentApp.canvas.onodegrid.ONodeTypes[oldName]
        beadBox.parentApp.canvas.onodegrid.ONodeTypes[newName] = newBeadType
        
        del self.beadBoxList[oldName]
        self.beadBoxList[newName] = beadBox

        for bbName in self.beadBoxList:
            bb = self.beadBoxList[bbName]
            if bb.beadType.name in newBondToTypes:
                bb.SetBackgroundColour("Yellow")
            else:
                bb.SetBackgroundColour(wx.NullColour)
            bb.Refresh()
            
        # Update the bead cycle with the new name
        for index, item in enumerate(beadBox.parentApp.canvas.onodegrid.beadCycle):
            if item == oldName:
                beadBox.parentApp.canvas.onodegrid.beadCycle[index] = newName
        beadBox.parentApp.DrawBeadCycle()
        
        # Update the assembly
        for i in range(beadBox.parentApp.canvas.onodegrid.minX, beadBox.parentApp.canvas.onodegrid.maxX+1):
            for j in range(beadBox.parentApp.canvas.onodegrid.minY, beadBox.parentApp.canvas.onodegrid.maxY+1):
                loc = beadBox.parentApp.canvas.onodegrid.GetLoc((i,j))
                if not loc is None:
                    if loc.type == oldName:
                        beadBox.parentApp.canvas.onodegrid.GetLoc((i,j)).type = newName

    def PickColor(self, event):
        if event.LeftDown() == True:
            box = event.EventObject.Parent
            if globalApp.useAdvancedColorPicker == False:
                dlg = wx.ColourDialog(self)
         
                # Ensure the full colour dialog is displayed, not the abbreviated version.
                dlg.GetColourData().SetChooseFull(True)
         
                if dlg.ShowModal() == wx.ID_OK:
                    data = dlg.GetColourData()
                    box.beadType.color = str(data.GetColour().Get())
                    box.colorButton.SetBitmapLabel(self.GetBeadButtonBitmap(box.beadType.color))
                    box.colorButton.Refresh()
                    globalApp.canvas.InitializeDrawing(globalApp.canvas.pdc)
                dlg.Destroy()
            else:
                cd = wx.ColourData()
                if type(box.beadType.color) is str:
                    cd.SetColour(wx.TheColourDatabase.Find(box.beadType.color))
                else:
                    cd.SetColour(box.beadType.color)
                dlg = CCD.CubeColourDialog(self, cd)
                if dlg.ShowModal() == wx.ID_OK:
                    colourData = dlg.GetColourData()
                    colorTuple = (colourData.GetColour().Red(), colourData.GetColour().Green(), colourData.GetColour().Blue(), colourData.GetColour().Alpha())
                    box.beadType.color = str(colorTuple)
                    #box.beadType.color = str(colourData.GetColour().Get())
                    box.colorButton.SetBitmapLabel(self.GetBeadButtonBitmap(box.beadType.color))
                    box.colorButton.Refresh()
                    globalApp.canvas.InitializeDrawing(globalApp.canvas.pdc)
                    self.Refresh()
                dlg.Destroy()

    def GetBeadButtonBitmap(self, color):
        w, h = 20, 20
        bmp = wx.EmptyBitmap(w, h)
        dc = wx.MemoryDC()
        dc.SelectObject(bmp)
        dc.Clear()
        pointBrush = self.parentApp.canvas.ColoredBrush(color)
        dc.SetBrush(pointBrush)
        dc.DrawCirclePoint((w/2,h/2), 10)
        dc.SelectObject(wx.NullBitmap)
        return bmp

#---------------------------------------------------------------------------


class OritatamiSimulatorApp(wx.App):
    def __init__(self):
        wx.App.__init__(self, redirect=False)
        self.showMessages = True
        self.messageQueue = []
        self.showDisplayUpdates = True
        self.useAdvancedColorPicker = False
        global globalApp
        globalApp = self


    def OnInit(self):

        wx.Log_SetActiveTarget(wx.LogStderr())

        self.SetAssertMode(assertMode)

        self.frame = wx.Frame(None, -1, "Oritatami Simulator", pos=(50,50), size=(400,400),
                         style=wx.DEFAULT_FRAME_STYLE, name="Oritatami Simulator")
        self.frame.SetIcon(wx.IconFromBitmap(wx.Bitmap("images/logo.png", wx.BITMAP_TYPE_PNG)))
        self.frame.CreateStatusBar(3)
        
        self.frame._mgr = wx.aui.AuiManager(self.frame)
        self.beadTypePanel          = None
        self.seqPanel               = None
        self.beadSequenceWin        = None
        self.messageWin             = None
        self.bpSize                 = (0,0)
        self.systemDefinitionFile   = None
        self.systemDefinitionFileFullPath = None
        self.seedDefinitionFile     = None
        self.assemblyDefinitionFile = None
        self.docFrame               = None
        
        provider = wx.SimpleHelpProvider()
        wx.HelpProvider.Set(provider)

        menuBar = wx.MenuBar()
        menu = wx.Menu()
        itemNew = menu.Append(-1, "New system", "New system")
        self.Bind(wx.EVT_MENU, self.OnNew, itemNew)
        menu.AppendSeparator()
        itemLSys = menu.Append(-1, "Load system definition", "Load system definition")
        self.Bind(wx.EVT_MENU, self.OnLoadSystem, itemLSys)
        item1 = menu.Append(-1, "&Load assembly\tCtrl-L", "Load assembly")
        self.Bind(wx.EVT_MENU, self.OnLoadAssembly, item1)
        itemR = menu.Append(-1, "Reload assembly", "Reload assembly")
        self.Bind(wx.EVT_MENU, self.OnReload, itemR)
        menu.AppendSeparator()
        itemR = menu.Append(-1, "&Reset to seed\tCtrl-R", "Reset system to the seed")
        self.Bind(wx.EVT_MENU, self.OnReset, itemR)
        menu.AppendSeparator()
        itemSBT = menu.Append(-1, "Save system definition", "Save oritatami system definition")
        self.Bind(wx.EVT_MENU, self.OnSaveSystemDef, itemSBT)
        itemSBTA = menu.Append(-1, "Save system definition as..", "Save oritatami system definition as..")
        self.Bind(wx.EVT_MENU, self.OnSaveSystemDefAs, itemSBTA)
        menu.AppendSeparator()
        itemSetSeed = menu.Append(-1, "Set current assembly as seed", "Set current assembly as seed assembly")
        self.Bind(wx.EVT_MENU, self.OnSetSeed, itemSetSeed)
        itemSS = menu.Append(-1, "Save seed", "Save seed assembly")
        self.Bind(wx.EVT_MENU, self.OnSaveSeed, itemSS)
        itemSSA = menu.Append(-1, "Save seed as..", "Save seed assembly as..")
        self.Bind(wx.EVT_MENU, self.OnSaveSeedAs, itemSSA)
        menu.AppendSeparator()
        itemSA = menu.Append(-1, "Save current assembly", "Save current assembly")
        self.Bind(wx.EVT_MENU, self.OnSaveAssembly, itemSA)
        itemSAA = menu.Append(-1, "Save current assembly as..", "Save current assembly as..")
        self.Bind(wx.EVT_MENU, self.OnSaveAssemblyAs, itemSAA)
        menu.AppendSeparator()
        item2 = menu.Append(-1, "&Save system definition and seed\tCtrl-S", "Save system definition and seed assembly")
        self.Bind(wx.EVT_MENU, self.OnSaveAll, item2)
        menu.AppendSeparator()
        item3 = menu.Append(-1, "E&xit\tCtrl-X", "Exit simulator")
        self.Bind(wx.EVT_MENU, self.OnExitApp, item3)
        menuBar.Append(menu, "&File")

        menuCtrl = wx.Menu()
        itemBP = menuCtrl.Append(-1, "&Breakpoints\tCtrl-B", "Set breakpoints")
        self.Bind(wx.EVT_MENU, self.OnBreakpoints, itemBP)
        menuCtrl.AppendSeparator()
        itemStep = menuCtrl.Append(-1, "&Forward\tCtrl-F", "Forward one step")
        self.Bind(wx.EVT_MENU, self.OnForward, itemStep)
        itemFF = menuCtrl.Append(-1, "&Go\tCtrl-G", "Fast forward")
        self.Bind(wx.EVT_MENU, self.OnGo, itemFF)
        itemQuit = menuCtrl.Append(-1, "&Quit\tCtrl-Q", "Quit fast forward")
        self.Bind(wx.EVT_MENU, self.OnQuit, itemQuit)
        menuBar.Append(menuCtrl, "&Control")

        self.menuView = wx.Menu()
        self.itemShowMess = self.menuView.Append(-1, "Show message window", "Toggle display of message window", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnShowMessageWin, self.itemShowMess)
        self.menuView.Check(self.itemShowMess.GetId(), True)
        self.itemShowBTEdit = self.menuView.Append(-1, "Show bead type editor window", "Toggle display of bead type editor window", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnShowBTEditWin, self.itemShowBTEdit)
        self.menuView.Check(self.itemShowBTEdit.GetId(), False)
        self.itemShowSeq = self.menuView.Append(-1, "Show bead sequence window", "Toggle display of bead sequence window", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnShowSequenceWin, self.itemShowSeq)
        self.menuView.Check(self.itemShowSeq.GetId(), True)
        self.menuView.AppendSeparator()
        itemSTC = self.menuView.Append(-1, "&Center\tCtrl+C", "Scroll to center")
        self.Bind(wx.EVT_MENU, self.OnCenter, itemSTC)
        self.menuView.AppendSeparator()
        itemZIn = self.menuView.Append(-1, "Zoom in\tCtrl-+", "Zoom in")
        self.Bind(wx.EVT_MENU, self.OnZoomIn, itemZIn)
        itemZOut = self.menuView.Append(-1, "Zoom out\tCtrl--", "Zoom out")
        self.Bind(wx.EVT_MENU, self.OnZoomOut, itemZOut)
        self.menuView.AppendSeparator()
        itemTG = self.menuView.Append(-1, "&Toggle grid lines\tCtrl-T", "Toggle display of grid lines")
        self.Bind(wx.EVT_MENU, self.OnToggleGrid, itemTG)
        self.itemTM = self.menuView.Append(-1, "Update &messages\tCtrl-M", "Toggle display of messages", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnMessageToggle, self.itemTM)
        self.menuView.Check(self.itemTM.GetId(), True)
        self.itemTD = self.menuView.Append(-1, "Update &display\tCtrl-D", "Toggle display of assembly updates", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnDisplayToggle, self.itemTD)
        self.menuView.Check(self.itemTD.GetId(), True)
        menuBar.Append(self.menuView, "&View")

        self.menuSettings = wx.Menu()
        itemSetDL = self.menuSettings.Append(-1, "Set delay length", "Set delay length system parameter")
        self.Bind(wx.EVT_MENU, self.OnSetDelayLength, itemSetDL)
        itemSetArity = self.menuSettings.Append(-1, "Set arity", "Set arity system parameter")
        self.Bind(wx.EVT_MENU, self.OnSetArity, itemSetArity)
        self.itemSetLWB = self.menuSettings.Append(-1, "Lock without binding", "Set lock without binding system parameter", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnSetLockWithoutBinding, self.itemSetLWB)
        self.menuSettings.Check(self.itemSetLWB.GetId(), False)
        self.itemSetBFFS = self.menuSettings.Append(-1, "Bind full floppy segment", "Set bind full floppy segment system parameter", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnSetBindFullFloppySegment, self.itemSetBFFS)
        self.menuSettings.Check(self.itemSetBFFS.GetId(), False)
        itemSetMFL = self.menuSettings.Append(-1, "Set max floppy length", "Set maximum floppy length to simulate")
        self.Bind(wx.EVT_MENU, self.OnSetMaxFloppyLength, itemSetMFL)
        self.menuSettings.AppendSeparator()
        self.itemAdvColor = self.menuSettings.Append(-1, "Use advanced color picker", "Use advanced color picker dialog", kind=wx.ITEM_CHECK)
        self.Bind(wx.EVT_MENU, self.OnAdvancedColorChecked, self.itemAdvColor)
        self.menuSettings.Check(self.itemAdvColor.GetId(), False)
        menuBar.Append(self.menuSettings, "&Settings")
    
        menu2 = wx.Menu()
        itemDoc = menu2.Append(-1, "Documentation", "Show documentation")
        self.Bind(wx.EVT_MENU, self.OnDoc, itemDoc)
        itemH = menu2.Append(-1, "&Help\tCtrl-H", "Show help")
        self.Bind(wx.EVT_MENU, self.OnHelp, itemH)
        menuBar.Append(menu2, "&Help")
        self.frame.SetMenuBar(menuBar)

        self.canvas = MyCanvas(self.frame, wx.ID_ANY, self.frame, wx.DefaultSize, self)
        self.frame._mgr.AddPane(self.canvas, wx.CENTER, 'Simulation Window')
        
        self.AddSequenceWindow()
        self.AddMessageWindow()
        
        self.frame.Bind(wx.EVT_CLOSE, self.OnCloseFrame)
        self.frame.Bind(wx.EVT_CHAR, self.canvas.OnChar)
        
        # Create toolbar
        tb1 = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        tb1.SetToolSeparation(20)
        tb1.SetToolBitmapSize(wx.Size(24,24))
        tb1.AddLabelTool(ORIT_ID_NEW, "New System", wx.Image("images/new.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="New System", longHelp="New oritatami system")
        tb1.AddLabelTool(ORIT_ID_LOAD, "Load System", wx.Image("images/load.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Load System", longHelp="Load oritatami system")
        tb1.AddLabelTool(ORIT_ID_SAVE, "Save System", wx.Image("images/save.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Save System", longHelp="Save oritatami system")
        tb1.AddLabelTool(ORIT_ID_RESET, "Reset System", wx.Image("images/reset.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Reset System", longHelp="Reset oritatami system")
        tb1.Realize()
        tb1.Bind(wx.EVT_TOOL, self.OnTool)

        self.frame._mgr.AddPane(tb1, wx.aui.AuiPaneInfo().
                          Name("File toolbar").Caption("File Toolbar").
                          ToolbarPane().Top())
        #tb1.AddSeparator()
        #tb1.AddSeparator()
        tb1 = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        tb1.SetToolSeparation(20)
        tb1.SetToolBitmapSize(wx.Size(24,24))
        tb1.AddLabelTool(ORIT_ID_FR, "Fast Rewind", wx.Image("images/fastleft.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Rewind", longHelp="Fast rewind")
        tb1.AddLabelTool(ORIT_ID_REV, "Rewind", wx.Image("images/left.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Rewind", longHelp="Rewind one step")
        tb1.AddLabelTool(ORIT_ID_STOP, "Pause", wx.Image("images/pause.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Pause", longHelp="Pause simulation")
        tb1.AddLabelTool(ORIT_ID_FOR, "Forward", wx.Image("images/right.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Forward", longHelp="Forward one step")
        tb1.AddLabelTool(ORIT_ID_FF, "Fast Forward", wx.Image("images/fastright.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Fast Forward", longHelp="Fast forward")
        tb1.AddSeparator()
        tb1.AddLabelTool(ORIT_ID_ZOOMIN, "Zoom In", wx.Image("images/zoomin.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom In", longHelp="Zoom in")
        tb1.AddLabelTool(ORIT_ID_ZOOMOUT, "Zoom Out", wx.Image("images/zoomout.xpm", wx.BITMAP_TYPE_XPM).ConvertToBitmap(), shortHelp="Zoom Out", longHelp="Zoom out")
        tb1.Realize()
        tb1.Bind(wx.EVT_TOOL, self.OnTool)

        self.frame._mgr.AddPane(tb1, wx.aui.AuiPaneInfo().
                          Name("Simulation toolbar").Caption("Simulation Toolbar").
                          ToolbarPane().Top())
        tb1 = wx.ToolBar(self.frame, -1, wx.DefaultPosition, wx.DefaultSize, wx.TB_DOCKABLE | wx.TB_FLAT)
        tb1.SetToolSeparation(10)
        #tb1.SetToolPacking(10)
        #tb1.SetToolBitmapSize(wx.Size(16,16))
        btnImg = wx.Image("images/pan-zoom.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(45, 25)
        tb1.AddRadioTool(ORIT_ID_MODE_PAN_ZOOM, btnImg.ConvertToBitmap(), shortHelp="Pan (left mouse)/zoom (right mouse)", longHelp="Pan (left mouse) and zoom (right mouse)")
        btnImg = wx.Image("images/add-remove-bead.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(25, 15)
        tb1.AddRadioTool(ORIT_ID_MODE_DRAW_BEAD, btnImg.ConvertToBitmap(), shortHelp="Add beads (left mouse)/remove beads (right mouse)", longHelp="Add beads (left mouse) and remove beads (right mouse)")
        btnImg = wx.Image("images/add-remove-link.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(60, 15)
        tb1.AddRadioTool(ORIT_ID_MODE_DRAW_LINK, btnImg.ConvertToBitmap(), shortHelp="Add links (left mouse)/remove links (right mouse)", longHelp="Add links (left mouse)/remove links (right mouse)")
        btnImg = wx.Image("images/add-remove-bond.png", wx.BITMAP_TYPE_PNG)
        btnImg.Rescale(60, 15)
        tb1.AddRadioTool(ORIT_ID_MODE_DRAW_BOND, btnImg.ConvertToBitmap(), shortHelp="Add bonds (left mouse)/remove bonds (right mouse)", longHelp="Add bonds (left mouse)/remove bonds (right mouse)")
        tb1.Realize()
        tb1.Bind(wx.EVT_TOOL, self.OnTool)

        self.frame._mgr.AddPane(tb1, wx.aui.AuiPaneInfo().
                          Name("Mouse mode toolbar").Caption("Mouse Mode Toolbar").
                          ToolbarPane().Top())
        

        # so set the frame to a good size for showing stuff
        self.frame.SetSize((1200, 800))
        self.frame.SetFocus()
        self.window = self.frame

        self.SetTopWindow(self.frame)
        #wx.Log_SetActiveTarget(wx.LogStderr())
        #wx.Log_SetTraceMask(wx.TraceMessages)
        
        self.AddBeadTypeWindow()
        
        # tell the manager to 'commit' all the changes just made
        self.frame._mgr.Update()
        #self.frame._mgr.OnFloatingPaneClosed()

        self.frame.Maximize()
        self.frame.Show(True)

        return True

    def AddMessageWindow(self):
        self.ctrlPanel = wx.Panel(self.frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(300,100), style = wx.TAB_TRAVERSAL)
        self.messageWin = rt.RichTextCtrl(self.ctrlPanel, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,70))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        self.messageWin.SetEditable(False)
        sz = wx.BoxSizer(wx.HORIZONTAL)
        sz.Add(self.messageWin,20,wx.EXPAND, 20)
        self.ctrlPanel.SetSizerAndFit(sz)
        self.frame._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseMessagePane)
        self.frame._mgr.AddPane(self.ctrlPanel, wx.BOTTOM, 'Message Panel')
        self.frame._mgr.Update()
        
    def AddSequenceWindow(self):
        self.seqPanel = wx.Panel(self.frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(300,100), style = wx.TAB_TRAVERSAL)
        
        sz = wx.BoxSizer(wx.HORIZONTAL)
        leftSizer = wx.BoxSizer(wx.VERTICAL)
        self.applyButton  = wx.Button(self.seqPanel, wx.ID_ANY, "Apply", size=(55,25))
        self.applyButton.Bind(wx.EVT_MOUSE_EVENTS, self.OnSequenceApply)
        leftSizer.Add(self.applyButton)
        self.setCurrButton  = wx.Button(self.seqPanel, wx.ID_ANY, "Current", size=(55,25))
        self.setCurrButton.Bind(wx.EVT_MOUSE_EVENTS, self.OnSetCurrentBead)
        #MJPTODO: figure out how to display mouse over text
        self.setCurrButton.SetHelpText("MJPTODO: add help text")
        leftSizer.Add(self.setCurrButton)
        self.resetButton  = wx.Button(self.seqPanel, wx.ID_ANY, "Reset", size=(55,25))
        self.resetButton.Bind(wx.EVT_MOUSE_EVENTS, self.OnSequenceReset)
        leftSizer.Add(self.resetButton)
        sz.Add(leftSizer)
    
        self.beadSequenceWin = rt.RichTextCtrl(self.seqPanel, wx.ID_ANY, pos=wx.DefaultPosition, size=(500,70))#, style=wx.NOSCROLL|wx.HSCROLL|wx.NO_BORDER)
        #self.beadSequenceWin.SetEditable(False)
        sz.Add(self.beadSequenceWin, 20, wx.EXPAND, 20)
        self.seqPanel.SetSizerAndFit(sz)
        self.frame._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseSequencePane)
        self.frame._mgr.AddPane(self.seqPanel, wx.BOTTOM, 'Bead Sequence Panel')
        self.frame._mgr.Update()
        
    def OnSetCurrentBead(self, event):
        if event.LeftDown():
            if not self.beadSequenceWin.HasSelection():
                globalApp.ShowErrorDialog("You must select a bead name to make it the current location")
                return
            else:
                selRange = self.beadSequenceWin.GetSelectionRange()
                selStart = selRange.GetStart()
                selEnd = selRange.GetEnd()
                selText = self.beadSequenceWin.GetBuffer().GetTextForRange((selStart, selEnd-1))
                if "," in selText:
                    globalApp.ShowErrorDialog("You must select only a single bead name to make it the current location")
                    return
                
                # Compare string in window against sequence currently being used and don't allow if different
                cycleStr = self.GetStringFromBeadCycle()
                winStr = str(self.beadSequenceWin.GetValue())
                if not cycleStr == winStr:
                    globalApp.ShowErrorDialog("The bead cycle currently in the text box differs from that used by the\nsimulator. Press 'Apply' to make it current or 'Reset' to revert first.")
                    return

                pos   = 0
                count = 0
                for char in cycleStr:
                    if pos == selStart:
                        break
                    if char == ',':
                        count = count + 1
                    pos = pos + 1
                
                # Looks good, update the current position and redraw the text
                self.canvas.onodegrid.currCyclePos = count
                self.DrawBeadCycle()
                self.canvas.SetFocus()
                
    def GetStringFromBeadCycle(self):
        outStr = ""
        for x in self.canvas.onodegrid.beadCycle:
            if len(outStr) > 0:
                outStr = outStr + ","
            outStr = outStr + x
        return outStr
        
    def OnSequenceApply(self, event):
        if event.LeftDown():
            cycleStr = self.GetStringFromBeadCycle()
            winStr = str(self.beadSequenceWin.GetValue()).strip()
            if not cycleStr == winStr:
                winArr = winStr.split(",")
                notDefined = []
                for bt in winArr:
                    if " " in bt:
                        globalApp.ShowErrorDialog("Bead type '" + bt + "' is invalid.  Spaces are not allowed.")
                        return
                    elif not bt in self.canvas.onodegrid.ONodeTypes:
                        notDefined.append(bt)
                if len(notDefined) > 0:
                    if len(notDefined) == 1:
                        globalApp.ShowErrorDialog("Bead type " + str(notDefined[0]) + " is not defined.")
                    else:
                        globalApp.ShowErrorDialog("Bead types " + str(notDefined) + " are not defined.")
                    return
                self.canvas.onodegrid.beadCycle = winArr
                globalApp.SetOutputMessage("The primary structure has been set to " + str(self.canvas.onodegrid.beadCycle))
                self.canvas.SetFocus()
        
    def OnSequenceReset(self, event):
        if event.LeftDown():
            #dlg = ResetBeadCycleDialog(self.beadSequenceWin, wx.ID_ANY, "Are you sure you want to reset the sequence and discard your changes?", caption="Confirm reset")
            dlg = ResetBeadCycleDialog(self.beadSequenceWin, wx.ID_ANY)
            dlg.CenterOnScreen()
            retId = dlg.ShowModal()
            if retId == ORIT_ID_RESET_BEADSEQ:
                dlg.Destroy()
                self.canvas.onodegrid.currCyclePos = 0
                self.DrawBeadCycle()
            elif retId == ORIT_ID_RELOAD_BEADSEQ:
                dlg.Destroy()
                self.canvas.onodegrid.beadCycle = self.canvas.onodegrid.origBeadCycle
                self.canvas.onodegrid.currCyclePos = 0
                self.DrawBeadCycle()
            else:
                dlg.Destroy()
            self.canvas.SetFocus()

    def OnTool(self, evt):
        if evt.GetId() == ORIT_ID_MODE_PAN_ZOOM:
            self.SetMouseMode(ORIT_ID_MODE_PAN_ZOOM)
        elif evt.GetId() == ORIT_ID_MODE_DRAW_BEAD:
            self.SetMouseMode(ORIT_ID_MODE_DRAW_BEAD)
        elif evt.GetId() == ORIT_ID_MODE_DRAW_LINK:
            self.SetMouseMode(ORIT_ID_MODE_DRAW_LINK)
        elif evt.GetId() == ORIT_ID_MODE_DRAW_BOND:
            self.SetMouseMode(ORIT_ID_MODE_DRAW_BOND)
        elif evt.GetId() == ORIT_ID_FR:
            globalApp.ShowErrorDialog("Sorry, this is not implemented yet")
        elif evt.GetId() == ORIT_ID_REV:
            globalApp.ShowErrorDialog("Sorry, this is not implemented yet")
        elif evt.GetId() == ORIT_ID_STOP:
            self.canvas.DoQuit()
        elif evt.GetId() == ORIT_ID_FOR:
            self.canvas.DoForwardStep()
        elif evt.GetId() == ORIT_ID_FF:
            self.canvas.DoGo()
        elif evt.GetId() == ORIT_ID_NEW:
            self.canvas.DoNew()
        elif evt.GetId() == ORIT_ID_LOAD:
            self.canvas.DoLoad(bAssembly=True, bSystem=True)
        elif evt.GetId() == ORIT_ID_RESET:
            self.canvas.Reset()
        elif evt.GetId() == ORIT_ID_SAVE:
            self.canvas.DoSaveAll()
        elif evt.GetId() == ORIT_ID_ZOOMIN:
            self.canvas.ZoomThenRecenter(True)
        elif evt.GetId() == ORIT_ID_ZOOMOUT:
            self.canvas.ZoomThenRecenter(False)

    def OnCloseMessagePane(self, event):
        if event.GetPane().window == self.ctrlPanel:
            self.menuView.Check(self.itemShowMess.GetId(), False)
        event.Skip()
        
    def OnShowMessageWin(self, event):
        pane = self.frame._mgr.GetPane(self.ctrlPanel)
        if self.itemShowMess.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()
        
    def OnCloseSequencePane(self, event):
        if event.GetPane().window == self.seqPanel:
            self.menuView.Check(self.itemShowSeq.GetId(), False)
        event.Skip()
        
    def OnShowSequenceWin(self, event):
        if self.seqPanel is None:
            self.AddSequenceWindow()
        pane = self.frame._mgr.GetPane(self.seqPanel)
        if self.itemShowSeq.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()
        
    def OnCloseBTEditPane(self, event):
        if event.GetPane().window == self.beadTypePanel:
            self.menuView.Check(self.itemShowBTEdit.GetId(), False)
        event.Skip()
        
    def OnShowBTEditWin(self, event):
        if self.beadTypePanel is None:
            self.AddBeadTypeWindow()
        pane = self.frame._mgr.GetPane(self.beadTypePanel)
        if self.itemShowBTEdit.IsChecked():
            pane.Show()
        else:
            pane.Hide()
        self.frame._mgr.Update()
        
    def OnSetDelayLength(self, event):
        dlg = wx.TextEntryDialog(self.frame, "Delay length:", caption="Set delay length value", defaultValue=str(self.canvas.onodegrid.delayLength))
        if dlg.ShowModal() == wx.ID_OK:
            try:
                self.canvas.onodegrid.delayLength = int(dlg.GetValue())
            except ValueError:
                globalApp.ShowErrorDialog("Delay length must be an integer value.")
        dlg.Destroy()
        
    def OnSetMaxFloppyLength(self, event):
        dlg = wx.TextEntryDialog(self.frame, "Maximum floppy length:", caption="Set maximum floppy length value", defaultValue=str(self.canvas.onodegrid.maxFloppyLength))
        if dlg.ShowModal() == wx.ID_OK:
            try:
                self.canvas.onodegrid.maxFloppyLength = int(dlg.GetValue())
            except ValueError:
                globalApp.ShowErrorDialog("Maximum floppy length must be an integer value.")
        dlg.Destroy()

    def OnSetArity(self, event):
        dlg = wx.TextEntryDialog(self.frame, "Arity:", caption="Set arity value", defaultValue=str(self.canvas.onodegrid.arity))
        if dlg.ShowModal() == wx.ID_OK:
            try:
                self.canvas.onodegrid.arity = int(dlg.GetValue())
            except ValueError:
                globalApp.ShowErrorDialog("Arity must be an integer value.")
        dlg.Destroy()
        
    def OnSetLockWithoutBinding(self, event):
        if self.itemSetLWB.IsChecked():
            self.canvas.onodegrid.lockWithoutBinding = True
        else:
            self.canvas.onodegrid.lockWithoutBinding = False
            
    def OnSetBindFullFloppySegment(self, event):
        if self.itemSetBFFS.IsChecked():
            self.canvas.onodegrid.bindFullFloppySegment = True
        else:
            self.canvas.onodegrid.bindFullFloppySegment = False
        
    def OnAdvancedColorChecked(self, event):
        if self.itemShowMess.IsChecked():
            globalApp.useAdvancedColorPicker = True
        else:
            globalApp.useAdvancedColorPicker = False
        
    def SetMouseMode(self, mode):
        if not mode == self.canvas.mouseMode:
            self.canvas.linkStart = None
            self.canvas.linkEnd = None
            if (self.canvas.posSelected is not None) or (self.canvas.negSelected is not None):
                self.canvas.posSelected = None
                self.canvas.negSelected = None
                self.canvas.InitializeDrawing(self.canvas.pdc)
        if mode == ORIT_ID_MODE_PAN_ZOOM:
            self.canvas.mouseMode = ORIT_MODE_PAN_ZOOM
        elif mode == ORIT_ID_MODE_DRAW_BEAD:
            self.canvas.mouseMode = ORIT_MODE_DRAW_BEAD
            if self.beadTypePanel is None:
                self.AddBeadTypeWindow()
            pane = self.frame._mgr.GetPane(self.beadTypePanel)
            self.menuView.Check(self.itemShowBTEdit.GetId(), True)
            pane.Show()
            self.frame._mgr.Update()
        elif mode == ORIT_ID_MODE_DRAW_LINK:
            self.canvas.mouseMode = ORIT_MODE_DRAW_LINK
        elif mode == ORIT_ID_MODE_DRAW_BOND:
            self.canvas.mouseMode = ORIT_MODE_DRAW_BOND
            
    def AddBeadTypeWindow(self):
        bIsNew = False
        if self.beadTypePanel == None:
            bIsNew = True
            self.beadTypePanel = scrolled.ScrolledPanel(self.frame, wx.ID_ANY, pos=wx.DefaultPosition, size=(400,150), style = wx.TAB_TRAVERSAL)
        else:
            oldSizer = self.beadTypePanel.GetSizer()
            for bb in self.beadBoxes:
                self.beadBoxes[bb].Destroy()
            oldSizer.Destroy()
        
        #beadTypePanelSizer = wx.BoxSizer(wx.VERTICAL)
        #btNum = len(self.canvas.onodegrid.ONodeTypes) + 1
        beadTypePanelSizer = wx.GridSizer(cols=1, hgap=5, vgap=5)
        
        #beadTypePanelSizer = wx.FlexGridSizer(0,0,5,5)
        #beadTypePanelSizer.SetFlexibleDirection(wx.HORIZONTAL)
        #beadTypePanelSizer.SetNonFlexibleGrowMode(wx.FLEX_GROWMODE_ALL)
        #beadTypePanelSizer.AddGrowableRow(0)
        
        #self.beadTypePanel.SetVirtualSize((1000,1000))
        
        self.beadBoxes = dict()
        bb = OritatamiBeadBox(None, self.beadTypePanel, beadTypePanelSizer, self, self.beadBoxes)
        self.beadBoxes[""] = bb

        for beadTypeName in self.canvas.onodegrid.ONodeTypes:
            beadType = self.canvas.onodegrid.ONodeTypes[beadTypeName]
            bb = OritatamiBeadBox(beadType, self.beadTypePanel, beadTypePanelSizer, self, self.beadBoxes)
            self.beadBoxes[beadTypeName] = bb
        beadTypePanelSizer.Layout()
        self.beadTypePanel.SetSizerAndFit(beadTypePanelSizer)
        
        if bIsNew == True:
            self.beadTypePanel.Bind(wx.EVT_MOUSE_EVENTS, self.OnBeadPanelClick)
            self.beadTypePanel.Bind(wx.EVT_SIZE, self.OnBeadPanelResize)
            self.frame._mgr.Bind(wx.aui.EVT_AUI_PANE_CLOSE, self.OnCloseBTEditPane)
            self.frame._mgr.AddPane(self.beadTypePanel, wx.LEFT, 'Bead Type Window')
        self.frame._mgr.Update()
        self.beadTypePanel.SetupScrolling()
        self.menuView.Check(self.itemShowBTEdit.GetId(), True)
        self.bpSize = self.beadTypePanel.GetSize()
        
    def OnBeadTypeDelete(self, event):
        if event.LeftDown():
            box = event.EventObject.Parent
            bname = str(box.editName.GetValue())
            dlg = wx.MessageDialog(self.beadTypePanel, "Are you sure you want to delete bead type '" + bname + "'?", caption="Confirm delete")
            dlgRet = dlg.ShowModal()
            dlg.Destroy()
            if dlgRet == wx.ID_OK:
                beadType = box.beadType

                # Delete the bead type from the assembly
                removeList = []
                for i in range(self.canvas.onodegrid.minX, self.canvas.onodegrid.maxX+1):
                    for j in range(self.canvas.onodegrid.minY, self.canvas.onodegrid.maxY+1):
                        loc = self.canvas.onodegrid.GetLoc((i, j))
                        if not loc is None:
                            if loc.type == beadType.name:
                                self.canvas.onodegrid.RemoveNode((i, j))
                                removeList.append((i, j))
                self.canvas._lock.acquire()
                self.canvas.RemoveQueue = self.canvas.RemoveQueue + removeList
                self.canvas._lock.release()
                self.canvas.DrawUpdate(self.canvas.pdc)
                self.canvas.Refresh()
                
                # Delete the bead type from the bead cycle
                newBeadCycle = []
                bDidDelete = False
                for b in self.canvas.onodegrid.beadCycle:
                    if not b == beadType.name:
                        newBeadCycle.append(b)
                        bDidDelete = True
                if bDidDelete == True:
                    self.canvas.onodegrid.beadCycle = newBeadCycle
                    self.canvas.onodegrid.InitializeFloppySegment()
                    self.DrawBeadCycle()
                
                # Delete the bead type box from the editor
                panelSizer = box.parentPanelSizer
                bbSizer = box.beadBoxSizer
                panelSizer.Hide(bbSizer)
                panelSizer.Remove(bbSizer)
                panelSizer.Layout()
                
                # Delete the bead type from the simulator's set
                del self.canvas.onodegrid.ONodeTypes[beadType.name]
        
    def OnBeadPanelResize(self, event):
        bp = event.EventObject
        w,h = bp.GetSize()
        if not (w,h) == self.bpSize:
            self.bpSize = (w,h)
            rowMod = w % 200
            numCols = int(w / 200)
            if rowMod > 50:
                numCols = numCols + 1
            if numCols < 1:
                numCols = 1
            #print("old size: " + str(self.bpSize) + ", new size: " + str((w,h)) + ", new num cols: " + str(numCols))
            sizer = bp.GetSizer()
            sizer.SetCols(numCols)
            #sizer.Layout()
            #self.frame._mgr.Update()
            
        
    def OnBeadPanelClickStatText(self, event):
        if event.LeftDown() == True:
            box = event.EventObject.Parent
            self.SelectCurrentDrawingBead(box)

    def OnBeadPanelClick(self, event):
        if event.LeftDown() == True:
            x = event.GetX()
            y = event.GetY()
            children = self.beadTypePanel.GetChildren()
            for child in children:
                rect = child.GetRect()
                if (x >= rect[0]) and (x <= (rect[0]+rect[2])) and (y >= rect[1]) and (y <= (rect[1]+rect[3])):
                    self.SelectCurrentDrawingBead(child)
        
    def SelectCurrentDrawingBead(self, clickedBox):
        if clickedBox.beadType.name == "":
            return
        if clickedBox.beadType == self.canvas.drawingModeBeadType:
            for bt in clickedBox.beadType.bondToTypes:
                self.beadBoxes[bt].SetBackgroundColour(wx.NullColour)
                self.beadBoxes[bt].Refresh()
            self.canvas.drawingModeBeadType = None
            clickedBox.SetBackgroundColour(wx.NullColour)
        else:
            if self.canvas.drawingModeBeadType is not None:
                self.beadBoxes[self.canvas.drawingModeBeadType.name].SetBackgroundColour(wx.NullColour)
                self.beadBoxes[self.canvas.drawingModeBeadType.name].Refresh()
                for bt in self.canvas.drawingModeBeadType.bondToTypes:
                    self.beadBoxes[bt].SetBackgroundColour(wx.NullColour)
                    self.beadBoxes[bt].Refresh()
            for bt in clickedBox.beadType.bondToTypes:
                self.beadBoxes[bt].SetBackgroundColour("Yellow")
                self.beadBoxes[bt].Refresh()
            self.canvas.drawingModeBeadType = clickedBox.beadType
            clickedBox.SetBackgroundColour("Grey")
        clickedBox.Refresh()
        
    def DrawBeadCycle(self):
        if (not self.showMessages):# or (self.beadSequenceWin == None):
            return
        
        self.beadSequenceWin.Freeze()
        self.beadSequenceWin.Clear()
        #outStr = str(self.canvas.onodegrid.beadCycle[0:self.canvas.onodegrid.currCyclePos]) + "X" + str(self.canvas.onodegrid.beadCycle[self.canvas.onodegrid.currCyclePos:])
        
        outStr = ""
        pos = 0
        for x in self.canvas.onodegrid.beadCycle:
            if pos == self.canvas.onodegrid.currCyclePos:
                if len(outStr) > 0:
                    self.beadSequenceWin.WriteText(outStr + ",")

                self.beadSequenceWin.BeginBold()
                self.beadSequenceWin.WriteText(x)
                self.beadSequenceWin.EndBold()
                outStr = ""
            elif pos > 0:
                outStr = outStr + "," + x
            else:
                outStr = outStr + x
            pos = pos + 1
        self.beadSequenceWin.WriteText(outStr)
        #self.beadSequenceWin.BeginFontSize(14)
        #self.beadSequenceWin.SetInsertionPointEnd()
        #self.beadSequenceWin.WriteText(outStr)
        #self.beadSequenceWin.SetInsertionPointEnd()
        #self.beadSequenceWin.EndFontSize()
        self.beadSequenceWin.Thaw()
        #self.beadSequenceWin.ScrollLines(self.beadSequenceWin.GetNumberOfLines()*5)
        
    def ResetBreakpoints(self):
        if len(self.canvas.breakpoints) > 0:
            self.canvas.currentBreakpoint = 0

    def SetOutputMessage(self, msg):
        if not self.showMessages:
            self.messageQueue.append(msg)
            return
        self.messageWin.SetInsertionPointEnd()
        self.messageWin.WriteText("\n" + msg)
        self.messageWin.SetInsertionPointEnd()
        self.messageWin.ScrollLines(self.messageWin.GetNumberOfLines()*5)
        #self.messageWin.Freeze()
        #self.messageWin.Clear()
        #self.messageWin.WriteText("\n" + msg)
        #self.messageWin.Thaw()
        #self.messageWin.ScrollLines(-1)

    def ShowErrorDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Error", style=wx.ICON_ERROR|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()
        
    def ShowWarningDialog(self, msg):
        dlg = wx.MessageDialog(self.frame, msg, caption="Warning", style=wx.ICON_WARNING|wx.OK)
        dlg.ShowModal()
        dlg.Destroy()
        
    def OnGo(self, event):
        self.canvas.DoGo()
        
    def OnQuit(self, event):
        self.canvas.DoQuit()

    def OnExitApp(self, evt):
        self.frame.Close(True)

    def OnNew(self, evt):
        self.canvas.DoNew()

    def OnLoadAssembly(self, evt):
        self.canvas.DoLoad(bAssembly=True, bSystem=False)
        
    def OnLoadSystem(self, evt):
        self.canvas.DoLoad(bAssembly=False, bSystem=True)

    def OnReload(self, evt):
        self.canvas.Reload()
        
    def OnReset(self, evt):
        self.canvas.Reset()

    def OnSaveSystemDef(self, event):
        self.canvas.DoSaveSystemDef()

    def OnSaveSystemDefAs(self, event):
        self.canvas.DoSaveSystemDef(bSaveAs=True)

    def OnSetSeed(self, event):
        self.canvas.DoSetSeed()

    def OnSaveSeed(self, event):
        self.canvas.DoSaveAssembly(bOnlySeed=True)

    def OnSaveSeedAs(self, event):
        self.canvas.DoSaveAssembly(bOnlySeed=True, bSaveAs=True)

    def OnSaveAssembly(self, event):
        self.canvas.DoSaveAssembly()

    def OnSaveAssemblyAs(self, event):
        self.canvas.DoSaveAssembly(bOnlySeed=False, bSaveAs=True)

    def OnSaveAll(self, event):
        self.canvas.DoSaveAll()
        
    def OnBreakpoints(self, event):
        self.canvas.SetBreakpoints()

    def OnForward(self, event):
        self.canvas.DoForwardStep()
        
    def OnCenter(self, event):
        self.canvas.ScrollToCoords((0,0))
        
    def OnZoomIn(self, event):
        self.canvas.ZoomThenRecenter(True)

    def OnZoomOut(self, event):
        self.canvas.ZoomThenRecenter(False)
        
    def OnToggleGrid(self, event):
        self.canvas.ToggleGrid()
    
    def OnMessageToggle(self, event):
        if self.itemTM.IsChecked():
            self.canvas.MessageToggle(True)
        else:
            self.canvas.MessageToggle(False)
        
    def OnDisplayToggle(self, event):
        if self.itemTD.IsChecked():
            self.canvas.DisplayToggle(True)
        else:
            self.canvas.DisplayToggle(False)

    def OnDoc(self, event):
        if self.docFrame is None:
            self.docFrame = DocumentationFrame(self)
        self.docFrame.SetFocus()

    def OnHelp(self, event):
        #info = wx.AboutDialogInfo()
        #info.SetName("Oritatami Simulator")
        #info.SetVersion("0.1.0")
        helpStr           = "                           Oritatami Simulator\n"
        helpStr = helpStr + "                               version 0.1.0\n\n"
        helpStr = helpStr + "    List of commands:\n\n"
        helpStr = helpStr + "      l : ('load') select oritatami system definition file to load\n"
        #helpStr = helpStr + "      s : save current assembly\n"
        helpStr = helpStr + "      r : ('reset') reset current system back to seed assembly\n"
        helpStr = helpStr + "      f : ('forward') simulate one step\n"
        helpStr = helpStr + "      g : ('go') begin fast forward simulation\n"
        helpStr = helpStr + "      q : ('quit') quit fast forward simulation\n"
        helpStr = helpStr + "      m : ('messges') toggle display of output messages\n"
        helpStr = helpStr + "      d : ('display') toggle updates to the display of the assembly\n"
        helpStr = helpStr + "      t : ('toggle') toggle display of grid lines\n"
        helpStr = helpStr + "      b : ('breakpoints') set breakpoints\n"
        helpStr = helpStr + "      +/-: zoom in/out\n"
        helpStr = helpStr + "      <left mouse click>:  zoom in on clicked location\n"
        helpStr = helpStr + "      <right mouse click>:  zoom out of clicked location\n"
        helpStr = helpStr + "      h : ('help') show this help dialog"
        helpStr = helpStr + "\n\n"
        #info.SetDescription(helpStr)
        #wx.AboutBox(info)
        helpPanel = wx.Panel(self.frame)
        helpText = wx.StaticText(helpPanel, label=helpStr)
        helpText.SetBackgroundColour((233,233,233))
        # add the panes to the manager
        self.frame._mgr.AddPane(helpText, wx.aui.AuiPaneInfo().Float().Caption("Oritatami Simulator Help"))
        self.frame._mgr.Update()
        

    def OnCloseFrame(self, evt):
        # Make sure the simulation thread is killed
        self.canvas.DoQuit()
        timer = wx.Timer(self)
        timer.Start(1000)
            #self.canvas.goThread.exit()

        if not self.docFrame is None:
            wx.PostEvent(self.docFrame, evt)

        # deinitialize the frame manager
        self.frame._mgr.UnInit()
        # delete the frame
        self.frame.Destroy()

#----------------------------------------------------------------------------

class ResetBeadCycleDialog(wx.Dialog):
    def __init__(
            self, parent, ID, size=wx.DefaultSize, pos=wx.DefaultPosition, style=wx.DEFAULT_DIALOG_STYLE):

        title = "Confirm bead sequence reset"
        # Instead of calling wx.Dialog.__init__ we precreate the dialog
        # so we can set an extra style that must be set before
        # creation, and then we create the GUI object using the Create
        # method.
        pre = wx.PreDialog()
        pre.Create(parent, ID, title, pos, size, style)

        # This next step is the most important, it turns this Python
        # object into the real wrapper of the dialog (instead of pre)
        # as far as the wxPython extension is concerned.
        self.PostCreate(pre)

        # Now continue with the normal construction of the dialog
        # contents
        sizer = wx.BoxSizer(wx.VERTICAL)

        label = wx.StaticText(self, -1, "Are you sure you want to reset the sequence and\ndiscard your changes?\nChoose 'Reset' to reset to last applied sequence\nChoose 'Reload' to reload original sequence\n")
        sizer.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        btnsizer = wx.StdDialogButtonSizer()
        
        
        self.resetBtn = wx.Button(self, wx.ID_YES, "Reset")
        self.resetBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        #self.resetBtn.SetDefault()
        btnsizer.AddButton(self.resetBtn)

        self.reloadBtn = wx.Button(self, wx.ID_NO, "Reload")
        self.reloadBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        btnsizer.AddButton(self.reloadBtn)
        
        self.cancelBtn = wx.Button(self, wx.ID_APPLY, "Cancel")
        self.cancelBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        btnsizer.AddButton(self.cancelBtn)

        btnsizer.Realize()

        sizer.Add(btnsizer, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALIGN_RIGHT|wx.ALL, 5)

        self.SetSizer(sizer)
        sizer.Fit(self)
        
    def OnDlgButton(self, event):
        if event.LeftDown():
            if event.EventObject == self.resetBtn:
                self.EndModal(ORIT_ID_RESET_BEADSEQ)
            if event.EventObject == self.reloadBtn:
                self.EndModal(ORIT_ID_RELOAD_BEADSEQ)
            if event.EventObject == self.cancelBtn:
                self.EndModal(ORIT_ID_CANCEL_BEADSEQ)
        
#----------------------------------------------------------------------------
class DontShowAgainDialog(wx.Dialog):
    def __init__(
            self, parent, ID, title, message, size=wx.DefaultSize, pos=wx.DefaultPosition, style=wx.DEFAULT_DIALOG_STYLE):

        self.bShowAgain = True
        # Instead of calling wx.Dialog.__init__ we precreate the dialog
        # so we can set an extra style that must be set before
        # creation, and then we create the GUI object using the Create
        # method.
        pre = wx.PreDialog()
        pre.Create(parent, ID, title, pos, size, style)

        # This next step is the most important, it turns this Python
        # object into the real wrapper of the dialog (instead of pre)
        # as far as the wxPython extension is concerned.
        self.PostCreate(pre)

        # Now continue with the normal construction of the dialog
        # contents
        sizer = wx.BoxSizer(wx.VERTICAL)

        label = wx.StaticText(self, -1, message)
        sizer.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        self.checkbox = wx.CheckBox(self, -1, "Do not show this warning again")
        sizer.Add(self.checkbox, 0, wx.ALIGN_RIGHT|wx.ALL, 5)

        btnsizer = wx.StdDialogButtonSizer()
        
        self.okBtn = wx.Button(self, wx.ID_OK, "OK")
        self.okBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        self.okBtn.SetDefault()
        btnsizer.AddButton(self.okBtn)
        
        self.cancelBtn = wx.Button(self, wx.ID_CANCEL, "Cancel")
        self.cancelBtn.Bind(wx.EVT_MOUSE_EVENTS, self.OnDlgButton)
        btnsizer.AddButton(self.cancelBtn)

        btnsizer.Realize()

        sizer.Add(btnsizer, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALIGN_RIGHT|wx.ALL, 5)

        self.SetSizer(sizer)
        sizer.Fit(self)
        
    def OnDlgButton(self, event):
        if event.LeftDown():
            bChecked = self.checkbox.GetValue()
            if event.EventObject == self.okBtn:
                if bChecked == True:
                    self.EndModal(ORIT_ID_OK_CHECK)
                else:
                    self.EndModal(ORIT_ID_OK_NOCHECK)
            if event.EventObject == self.cancelBtn:
                self.EndModal(ORIT_ID_CANCEL)
        
#----------------------------------------------------------------------------

class SomeCustomDialog(wx.Dialog):
    def __init__(
            self, parent, ID, size=wx.DefaultSize, pos=wx.DefaultPosition, 
            style=wx.DEFAULT_DIALOG_STYLE,
            useMetal=False,
            ):

        title = "Custom dialog"
        # Instead of calling wx.Dialog.__init__ we precreate the dialog
        # so we can set an extra style that must be set before
        # creation, and then we create the GUI object using the Create
        # method.
        pre = wx.PreDialog()
        pre.SetExtraStyle(wx.DIALOG_EX_CONTEXTHELP)
        pre.Create(parent, ID, title, pos, size, style)

        # This next step is the most important, it turns this Python
        # object into the real wrapper of the dialog (instead of pre)
        # as far as the wxPython extension is concerned.
        self.PostCreate(pre)

        # This extra style can be set after the UI object has been created.
        if 'wxMac' in wx.PlatformInfo and useMetal:
            self.SetExtraStyle(wx.DIALOG_EX_METAL)


        # Now continue with the normal construction of the dialog
        # contents
        sizer = wx.BoxSizer(wx.VERTICAL)

        label = wx.StaticText(self, -1, "This is a wx.Dialog")
        label.SetHelpText("This is the help text for the label")
        sizer.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        box = wx.BoxSizer(wx.HORIZONTAL)

        label = wx.StaticText(self, -1, "Field #1:")
        label.SetHelpText("This is the help text for the label")
        box.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        text = wx.TextCtrl(self, -1, "", size=(80,-1))
        text.SetHelpText("Here's some help text for field #1")
        box.Add(text, 1, wx.ALIGN_CENTRE|wx.ALL, 5)

        sizer.Add(box, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)

        box = wx.BoxSizer(wx.HORIZONTAL)

        label = wx.StaticText(self, -1, "Field #2:")
        label.SetHelpText("This is the help text for the label")
        box.Add(label, 0, wx.ALIGN_CENTRE|wx.ALL, 5)

        text = wx.TextCtrl(self, -1, "", size=(80,-1))
        text.SetHelpText("Here's some help text for field #2")
        box.Add(text, 1, wx.ALIGN_CENTRE|wx.ALL, 5)

        sizer.Add(box, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)

        line = wx.StaticLine(self, -1, size=(20,-1), style=wx.LI_HORIZONTAL)
        sizer.Add(line, 0, wx.GROW|wx.ALIGN_CENTER_VERTICAL|wx.RIGHT|wx.TOP, 5)

        btnsizer = wx.StdDialogButtonSizer()
        
        if wx.Platform != "__WXMSW__":
            btn = wx.ContextHelpButton(self)
            btnsizer.AddButton(btn)
        
        btn = wx.Button(self, wx.ID_OK)
        btn.SetHelpText("The OK button completes the dialog")
        btn.SetDefault()
        btnsizer.AddButton(btn)

        btn = wx.Button(self, wx.ID_CANCEL)
        btn.SetHelpText("The Cancel button cancels the dialog. (Cool, huh?)")
        btnsizer.AddButton(btn)
        btnsizer.Realize()

        sizer.Add(btnsizer, 0, wx.ALIGN_CENTER_VERTICAL|wx.ALL, 5)

        self.SetSizer(sizer)
        sizer.Fit(self)

#----------------------------------------------------------------------------

class DocumentationTB(wx.Treebook):
    def __init__(self, parent, idIn):
        wx.Treebook.__init__(self, parent, idIn, style=
                             wx.BK_DEFAULT
                             #wx.BK_TOP
                             #wx.BK_BOTTOM
                             #wx.BK_LEFT
                             #wx.BK_RIGHT
                            )
        
        self.windows = []
        # Now make a bunch of panels for the list book
        for fileName in DocumentationFileList:
            win = self.MakeDocPanel(fileName)
            self.AddPage(win, fileName)
            self.windows.append(win)
            #win = self.MakeDocPanel(colour)
            #self.AddSubPage(win, 'a sub-page')

        # This is a workaround for a sizing bug on Mac...
        #wx.FutureCall(100, self.AdjustSize)

        self.Bind(wx.EVT_TREEBOOK_PAGE_CHANGING, self.OnPageChanging)

    def AdjustSize(self):
        #print self.GetTreeCtrl().GetBestSize()
        self.GetTreeCtrl().InvalidateBestSize()
        self.SendSizeEvent()
        #print self.GetTreeCtrl().GetBestSize()
        

    def MakeDocPanel(self, fileName):
        p = wx.html.HtmlWindow(self) 
        if "gtk2" in wx.PlatformInfo: 
            p.SetStandardFonts() 
        p.LoadPage("docs/" + fileName + ".html")
        return p

    def OnPageChanging(self, event):
        selected = event.GetSelection()
        #print("new: " + str(selected))
        correctDoc = DocumentationFileList[selected]
        self.windows[selected].LoadPage("docs/" + correctDoc + ".html")
        event.Skip()

#----------------------------------------------------------------------------

class DocumentationFrame(wx.Frame):
    def __init__(self, app):
        wx.Frame.__init__(self, None, -1, "Oritatami Simulator Documentation", pos=(50,50), size=(400,400),
                         style=wx.DEFAULT_FRAME_STYLE, name="Oritatami Simulator Documentation")
        self.app = app
        self._mgr = wx.aui.AuiManager(self)
        self.SetIcon(wx.IconFromBitmap(wx.Bitmap("images/logo.png", wx.BITMAP_TYPE_PNG)))

        self.treewin = DocumentationTB(self, -1)
        self._mgr.AddPane(self.treewin, wx.CENTER, 'Documentation Window')
        
        self.Bind(wx.EVT_CLOSE, self.OnCloseDocFrame)
        self.CreateStatusBar(3)
        self.SetSize((1200, 800))
        self.SetFocus()

        self._mgr.Update()
        #self.Maximize()
        self.Show(True)
        
    def OnCloseDocFrame(self, evt):
        # deinitialize the frame manager
        self._mgr.UnInit()
        self.app.docFrame = None
        # delete the frame
        self.Destroy()

#----------------------------------------------------------------------------
#  Create the application and run it

app = OritatamiSimulatorApp()
app.MainLoop()

#---------------------------------------------------------------------------

